#  Copyright (c) 2023, Manfred Moitzi
#  License: MIT License

import pytest
import ezdxf
from ezdxf.entities import DXFEntity
from ezdxf.document import Drawing
from ezdxf import r12strict


def set_xdata(entity: DXFEntity):
    entity.set_xdata("my appid", [(1003, "my layer")])


def setup_drawing(doc: Drawing):
    doc.appids.add("my appid")
    doc.linetypes.add("my ltype", [0])
    doc.layers.add("my layer", linetype="my ltype")
    doc.styles.add("my style", font="arial.ttf")
    doc.dimstyles.add(
        "my dimstyle",
        dxfattribs={
            "dimblk": "my block 0",
            "dimblk1": "my block 1",
            "dimblk2": "my block 2",
        },
    )
    doc.ucs.add("my ucs")
    doc.views.add("my view")
    doc.viewports.add("my vport")

    doc.header["$CELTYPE"] = "my ltype"
    doc.header["$CLAYER"] = "my layer"
    doc.header["$DIMBLK"] = "my block 0"
    doc.header["$DIMBLK1"] = "my block 1"
    doc.header["$DIMBLK2"] = "my block 2"
    doc.header["$DIMSTYLE"] = "my dimstyle"
    doc.header["$TEXTSTYLE"] = "my style"
    doc.header["$UCSNAME"] = "my ucs"

    common_attribs = {"layer": "my layer", "linetype": "my ltype"}
    for name in ["my block 0", "my block 1", "my block 2"]:
        blk = doc.blocks.new(name)
        blk.add_point((0, 0), dxfattribs=common_attribs)

    msp = doc.modelspace()
    set_xdata(msp.block_record.block)

    msp.add_line((0, 0), (1, 0), dxfattribs=common_attribs)

    text = msp.add_text("any text", dxfattribs=common_attribs)
    text.dxf.style = "my style"
    set_xdata(text)

    insert = msp.add_blockref("my block 0", (0, 0), dxfattribs=common_attribs)
    insert.add_attrib("my attrib", "any value")

    msp.add_linear_dim(
        (0, 0), (0, 0), (1, 0), dimstyle="my dimstyle", dxfattribs=common_attribs
    )


@pytest.fixture(scope="module")
def doc():
    doc = ezdxf.new("R12")
    setup_drawing(doc)
    r12strict.translate_names(doc)
    return doc


def test_translated_header_vars(doc: Drawing):
    assert doc.header["$CLAYER"] == "MY_LAYER"
    assert doc.header["$CELTYPE"] == "MY_LTYPE"
    assert doc.header["$DIMBLK"] == "MY_BLOCK_0"
    assert doc.header["$DIMBLK1"] == "MY_BLOCK_1"
    assert doc.header["$DIMBLK2"] == "MY_BLOCK_2"
    assert doc.header["$DIMSTYLE"] == "MY_DIMSTYLE"
    assert doc.header["$TEXTSTYLE"] == "MY_STYLE"
    assert doc.header["$UCSNAME"] == "MY_UCS"


def test_translated_appid(doc: Drawing):
    my_appid = doc.appids.get("MY_APPID")
    assert my_appid.dxf.name == "MY_APPID"


def test_translated_ltype(doc: Drawing):
    my_ltype = doc.linetypes.get("MY_LTYPE")
    assert my_ltype.dxf.name == "MY_LTYPE"


def test_translated_layer(doc: Drawing):
    my_layer = doc.layers.get("MY_LAYER")
    assert my_layer.dxf.name == "MY_LAYER"
    assert my_layer.dxf.linetype == "MY_LTYPE"


def test_translated_text_style(doc: Drawing):
    my_style = doc.styles.get("MY_STYLE")
    assert my_style.dxf.name == "MY_STYLE"


def test_translated_dimstyle(doc: Drawing):
    my_dimstyle = doc.dimstyles.get("MY_DIMSTYLE")
    assert my_dimstyle.dxf.name == "MY_DIMSTYLE"
    assert my_dimstyle.dxf.dimblk == "MY_BLOCK_0"
    assert my_dimstyle.dxf.dimblk1 == "MY_BLOCK_1"
    assert my_dimstyle.dxf.dimblk2 == "MY_BLOCK_2"


def test_translated_ucs(doc: Drawing):
    my_ucs = doc.ucs.get("MY_UCS")
    assert my_ucs.dxf.name == "MY_UCS"


def test_translated_view(doc: Drawing):
    my_view = doc.views.get("MY_VIEW")
    assert my_view.dxf.name == "MY_VIEW"


def test_translated_vport(doc: Drawing):
    my_vport = doc.viewports.get("MY_VPORT")
    assert my_vport[0].dxf.name == "MY_VPORT"


def test_translated_common_attributes(doc: Drawing):
    for entity in doc.modelspace():
        assert entity.dxf.layer == "MY_LAYER"
        assert entity.dxf.linetype == "MY_LTYPE"


def test_translated_text_attributes(doc: Drawing):
    text = doc.modelspace().query("TEXT").first
    assert text.dxf.style == "MY_STYLE"


def test_translated_text_xdata(doc: Drawing):
    text = doc.modelspace().query("TEXT").first
    xdata = text.get_xdata("MY_APPID")
    assert xdata[0] == (1003, "MY_LAYER")


def test_translated_dimension_attributes(doc: Drawing):
    dimension = doc.modelspace().query("DIMENSION").first
    assert dimension.dxf.dimstyle == "MY_DIMSTYLE"


def test_translated_block_reference_attributes(doc: Drawing):
    insert = doc.modelspace().query("INSERT").first
    attribute = insert.attribs[0]
    assert attribute.dxf.tag == "MY_ATTRIB"


def test_translated_modelspace_xdata(doc: Drawing):
    msp = doc.modelspace()
    xdata = msp.block_record.block.get_xdata("MY_APPID")
    assert xdata[0] == (1003, "MY_LAYER")


@pytest.mark.parametrize("name", ["MY_BLOCK_0", "MY_BLOCK_1", "MY_BLOCK_0"])
def test_translated_block_layout_names(doc: Drawing, name: str):
    assert name in doc.blocks
    block_layout = doc.blocks.get(name)
    assert block_layout.name == name
    assert block_layout.block_record.dxf.name == name
    assert block_layout.block_record.block.dxf.name == name


def test_translated_block_content_attributes(doc: Drawing):
    assert "MY_BLOCK_0" in doc.blocks
    block_layout = doc.blocks.get("MY_BLOCK_0")
    for entity in block_layout:
        assert entity.dxf.layer == "MY_LAYER"
        assert entity.dxf.linetype == "MY_LTYPE"


def test_clean_xdata_from_table_entries():
    doc = ezdxf.new("R12")
    doc.appids.new("MY_ID")
    layer = doc.layers.new("MY_LAYER")
    layer.set_xdata("MY_ID", [(1000, "Test")])

    r12strict.clean(doc)
    assert layer.xdata is None, "expected XDATA to be deleted"


def test_remove_legacy_blocks():
    doc = ezdxf.new("R12")
    doc.blocks.new("$Model_Space")
    doc.blocks.new("$Paper_Space")

    r12strict.clean(doc)
    # block names are case-insensitive
    assert "$MODEL_SPACE" not in doc.blocks
    assert "$PAPER_SPACE" not in doc.blocks
    assert doc.block_records.has_entry("$Model_Space") is False
    assert doc.block_records.has_entry("$Paper_Space") is False


if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2023-2024, Manfred Moitzi
#  License: MIT License
import pytest
import math

from ezdxf.tools.clipping_portal import ClippingRect
from ezdxf.math import Vec2
from ezdxf import npshapes


class TestClippingRect:
    def test_clipping_points(self):
        clipper = ClippingRect([(-1, -1), (1, 1)])
        assert clipper.clip_point(Vec2(3, 3)) is None
        assert clipper.clip_point(Vec2(0.5, 0.5)).isclose((0.5, 0.5))  # type: ignore

    def test_clipping_lines(self):
        clipper = ClippingRect([(-1, -1), (1, 1)])
        cropped_segments = clipper.clip_line(Vec2(-3, 0), Vec2(3, 0))
        assert len(cropped_segments) == 1

        points = cropped_segments[0]
        assert len(points) == 2
        start, end = points
        assert start.isclose((-1, 0))
        assert end.isclose((1, 0))

    def test_clip_polyline(self):
        clipper = ClippingRect([(-2, -2), (2, 2)])
        polyline = npshapes.NumpyPoints2d(Vec2.list([(0, 1), (3, 0), (0, -1)]))
        polygons = clipper.clip_polygon(polyline)
        assert len(polygons) == 1
        assert len(polygons[0]) == 4


if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2022, Manfred Moitzi
#  License: MIT License

import pytest

from ezdxf.addons import MengerSponge


@pytest.fixture(scope="module")
def mesh():
    return MengerSponge().mesh()


def test_euler_characteristic_for_menger_sponge(mesh):
    """The euler characteristic does not work for non-convex meshes."""
    diag = mesh.diagnose()
    assert diag.euler_characteristic == -8


def test_edge_balance_for_menger_sponge(mesh):
    """ The is_edge_balance_broken property is not reliable for concave meshes
    each edge connects two faces
    """
    diag = mesh.diagnose()
    assert diag.is_edge_balance_broken is False


def test_menger_sponge_is_non_manifold(mesh):
    diag = mesh.diagnose()
    assert diag.is_manifold is True


def test_menger_sponge_face_count(mesh):
    assert len(mesh.faces) == 72


def test_menger_sponge_edge_count(mesh):
    assert len(mesh.vertices) == 64


if __name__ == '__main__':
    pytest.main([__file__])



# Copyright (c) 2010-2020 Manfred Moitzi
# License: MIT License
import pytest
from io import StringIO
from ezdxf.lldxf.tagwriter import TagWriter, TagCollector
from ezdxf.lldxf.types import DXFTag, DXFVertex


def setup_stream():
    stream = StringIO()
    tagwriter = TagWriter(stream)
    return stream, tagwriter


def test_write_tag2():
    s, t = setup_stream()
    t.write_tag2(0, "SECTION")
    result = s.getvalue()
    assert result == "  0\nSECTION\n"


def test_write_tag():
    s, t = setup_stream()
    t.write_tag(DXFTag(0, "SECTION"))
    result = s.getvalue()
    assert result == "  0\nSECTION\n"


def test_write_point_tag():
    s, t = setup_stream()
    t.write_tag(DXFVertex(10, (7.0, 8.0, 9.0)))
    result = s.getvalue()
    assert result == " 10\n7.0\n 20\n8.0\n 30\n9.0\n"


def test_write_str():
    s, t = setup_stream()
    t.write_str(" 10\n7.0\n 20\n8.0\n 30\n9.0\n")
    result = s.getvalue()
    assert result == " 10\n7.0\n 20\n8.0\n 30\n9.0\n"


def test_write_anything():
    s, t = setup_stream()
    t.write_str("... writes just any nonsense ...")
    result = s.getvalue()
    assert result == "... writes just any nonsense ..."


class TestTagCollector:
    @pytest.fixture
    def t(self):
        return TagCollector()

    def test_write_tag2(self, t):
        t.write_tag2(0, "SECTION")
        assert t.tags[0] == (0, "SECTION")

    def test_write_tag(self, t):
        t.write_tag(DXFTag(0, "SECTION"))
        assert t.tags[0] == (0, "SECTION")

    def test_write_point_tag(self, t):
        t.write_tag(DXFVertex(10, (7.0, 8.0, 9.0)))
        assert t.tags[0] == (10, 7.0)
        assert t.tags[1] == (20, 8.0)
        assert t.tags[2] == (30, 9.0)

    def test_write_str(self, t):
        t.write_str(" 10\n7.0\n 20\n8.0\n 30\n9.0\n")
        assert t.tags[0] == (10, 7.0)
        assert t.tags[1] == (20, 8.0)
        assert t.tags[2] == (30, 9.0)



#  Copyright (c) 2025, Manfred Moitzi
#  License: MIT License
import pytest
from pathlib import Path

import ezdxf
from ezdxf import xref
from ezdxf.document import Drawing

DATA = Path(__file__).parent / "data"


@pytest.fixture(scope="module")
def doc1():
    return ezdxf.readfile(DATA / "issue_1359_1.dxf")


@pytest.fixture(scope="module")
def doc2():
    return ezdxf.readfile(DATA / "issue_1359_2.dxf")


@pytest.fixture(scope="module")
def mdoc(doc1: Drawing, doc2: Drawing):
    merged_doc = ezdxf.new(dxfversion=doc1.dxfversion)
    xref.load_modelspace(doc1, merged_doc, conflict_policy=xref.ConflictPolicy.KEEP)
    xref.load_modelspace(doc2, merged_doc, conflict_policy=xref.ConflictPolicy.KEEP)
    return merged_doc


def test_mdoc_contains_block_definition_1(mdoc: Drawing):
    assert "1" in mdoc.blocks


def test_merged_doc_contains_the_source_block_references(mdoc: Drawing):
    msp = mdoc.modelspace()
    assert len(msp) == 2
    assert all(e.dxftype() == "INSERT" for e in msp)


def test_copied_insert_entities_referencing_the_same_block(mdoc: Drawing):
    """Both INSERT entities should reference the BLOCK '1'."""
    msp = mdoc.modelspace()
    assert all(e.dxf.name == "1" for e in msp)


if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2021-2024, Manfred Moitzi
#  License: MIT License
from __future__ import annotations
from typing import TYPE_CHECKING, Iterator, Iterable, Any
from typing_extensions import Protocol, runtime_checkable
from ezdxf.query import EntityQuery

if TYPE_CHECKING:
    from ezdxf.entities import DXFGraphic, DXFEntity
    from ezdxf.entities.temporary_transform import TemporaryTransformation
    from ezdxf.math import Matrix44, AbstractBoundingBox


# Protocol implemented for:
# - DXFTagStorage
# - ACAD_PROXY_ENTITY
# - INSERT
# - DIMENSION
# - LEADER
# - MLEADER
# - MLINE
# - ProxyGraphic
# - DXFGraphicProxy (drawing add-on)
# - DXFTagStorage
# - ACAD_TABLE (table content loader)


@runtime_checkable
class SupportsVirtualEntities(Protocol):
    """The virtual entities protocol is used to disassemble complex entities
    into DXF primitives like LINE, ARC, ... as REQUIREMENT to render these
    entities. Which means the entity does not have :func:`ezdxf.path.make_path`
    support, except the text entities TEXT, ATTRIB and MTEXT.

    Optional DECONSTRUCTION of entities into DXF primitives like LWPOLYLINE
    into LINE and ARC entities is NOT the intended usage of this protocol!

    This protocol is for consistent internal usage and does not replace
    the :meth:`virtual_entities` methods!

    """

    def __virtual_entities__(self) -> Iterator[DXFGraphic]: ...


def virtual_entities(entity: SupportsVirtualEntities) -> Iterator[DXFGraphic]:
    if isinstance(entity, SupportsVirtualEntities):
        return entity.__virtual_entities__()
    else:
        raise TypeError(
            f"{type(entity)!r} does not support the __virtual_entities__ protocol"
        )


def query_virtual_entities(entity: SupportsVirtualEntities) -> EntityQuery:
    return EntityQuery(virtual_entities(entity))


@runtime_checkable
class ReferencedBlocks(Protocol):
    def __referenced_blocks__(self) -> Iterable[str]:
        """Returns the handles to the BLOCK_RECORD entities for all BLOCK
        definitions used by an entity.
        """


class SupportsTransform(Protocol):
    def transform(self, m: Matrix44) -> DXFGraphic:
        """Raises NotImplementedError() if transformation is not supported."""


_EMPTY_TUPLE: tuple = tuple()


def referenced_blocks(entity: DXFEntity) -> Iterable[str]:
    """Returns the handles to the BLOCK_RECORD entities for all BLOCK
    definitions used by an entity.
    """
    if isinstance(entity, ReferencedBlocks):
        return entity.__referenced_blocks__()
    else:
        return _EMPTY_TUPLE


class SupportsBoundingBox(Protocol):
    def bbox(self) -> AbstractBoundingBox: ...


@runtime_checkable
class SupportsTemporaryTransformation(Protocol):
    def temporary_transformation(self) -> TemporaryTransformation: ...


class SupportsMessages(Protocol):
    def notify(self, message_type: int, data: Any = None) -> None:
        """Internal messaging system."""



# Copyright (c) 2019 Manfred Moitzi
# License: MIT License
# Created 2019-02-13
import pytest

from ezdxf.entities.appdata import Reactors
from ezdxf.lldxf.const import (
    REACTOR_HANDLE_CODE,
    ACAD_REACTORS,
    APP_DATA_MARKER,
)
from ezdxf.lldxf.tags import Tags


class TagWriter:
    """Mockup"""

    def __init__(self):
        self.tags = []

    def write_tag2(self, code, value):
        self.tags.append((code, value))


def test_reactors_new():
    reactors = Reactors(["DDDD", "AA", "CCCC", "BBB"])
    assert len(reactors) == 4
    handles = reactors.get()
    # sorted handles
    assert handles[0] == "AA"
    assert handles[3] == "DDDD"

def test_copy_reactors():
    reactors0 = Reactors(["DDDD", "AA", "CCCC", "BBB"])
    reactors = reactors0.copy()
    assert reactors is not reactors0
    
    assert len(reactors) == 4
    handles = reactors.get()
    # sorted handles
    assert handles[0] == "AA"
    assert handles[3] == "DDDD"


def test_reactors_add():
    reactors = Reactors(["AA", "BBB", "CCCC"])
    reactors.add("AA")
    assert len(reactors) == 3, "do not add duplicates"
    reactors.add("0")
    assert len(reactors) == 4, "add unique handles"


def test_reactors_set():
    reactors = Reactors()
    assert len(reactors) == 0
    reactors.set(["0", "1", "2"])
    assert len(reactors) == 3
    reactors.set(["0"])
    assert len(reactors) == 1
    # delete all
    reactors.set([])
    assert len(reactors) == 0


def test_reactors_discard():
    reactors = Reactors(["AA", "BBB", "CCCC"])
    reactors.discard("AA")
    assert len(reactors) == 2

    # ignore non existing handles
    reactors.discard("abcd")
    assert len(reactors) == 2


def test_export_dxf():
    reactors = Reactors(["AA", "BBB", "CCCC"])
    tagwriter = TagWriter()
    reactors.export_dxf(tagwriter)
    tags = tagwriter.tags
    assert len(tags) == 5
    # sorted handles!
    assert tags[0] == (APP_DATA_MARKER, ACAD_REACTORS)
    assert tags[1] == (REACTOR_HANDLE_CODE, "AA")
    assert tags[2] == (REACTOR_HANDLE_CODE, "BBB")
    assert tags[3] == (REACTOR_HANDLE_CODE, "CCCC")
    assert tags[4] == (APP_DATA_MARKER, "}")


def test_from_tags():
    reactors = Reactors.from_tags(Tags.from_text(HANDLES))
    handles = reactors.get()
    assert len(handles) == 3
    assert handles[0] == "C000"
    assert handles[1] == "D000"
    assert handles[2] == "E000"


HANDLES = """102
{ACAD_REACTORS
330
D000
330
C000
330
E000
102
}
"""


if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2022-2024, Manfred Moitzi
#  License: MIT License

import pytest
from ezdxf.acis.api import load, export_sat, export_sab, ExportError
from ezdxf.acis import sat, sab, entities, hdr, const, mesh
from ezdxf.math import Matrix44
import math


def test_load_any_format(any_cube):
    bodies = load(any_cube)
    assert len(bodies) == 1


@pytest.fixture(scope="module")
def body(any_cube):
    return load(any_cube)[0]


class TestBody:
    def test_type_type(self, body):
        assert body.type == "body"

    def test_has_transform_attribute(self, body):
        assert body.transform.is_none is False

    def test_transform_attribute_was_loaded(self, body):
        m = body.transform.matrix
        assert m.get_row(3) == (388.5, 388.5, 388.5, 1.0)

    def test_has_wire_attribute(self, body):
        assert body.wire.is_none is True

    def test_get_all_lumps_as_a_list(self, body):
        bodies = body.lumps()
        assert len(bodies) == 1
        assert isinstance(bodies[0], entities.Lump)

    def test_append_lumps(self):
        body = entities.Body()
        assert len(body.lumps()) == 0
        body.append_lump(entities.Lump())
        assert len(body.lumps()) == 1
        body.append_lump(entities.Lump())
        assert len(body.lumps()) == 2


class TestLump:
    def test_lump_type(self, body):
        assert body.lump.type == "lump"

    def test_back_pointer_to_body(self, body):
        assert body.lump.body is body

    def test_has_no_next_lump(self, body):
        assert body.lump.next_lump.is_none is True

    def test_has_attribute_to_first_shell(self, body):
        assert body.lump.shell.is_none is False

    def test_get_all_shells_as_a_list(self, body):
        shells = body.lump.shells()
        assert len(shells) == 1
        assert isinstance(shells[0], entities.Shell)

    def test_append_shells(self):
        body = entities.Body()
        lump = entities.Lump()
        body.append_lump(lump)
        assert len(lump.shells()) == 0
        lump.append_shell(entities.Shell())
        assert len(lump.shells()) == 1
        lump.append_shell(entities.Shell())
        assert len(lump.shells()) == 2


class TestShell:
    @pytest.fixture(scope="class")
    def shell(self, body):
        return body.lump.shell

    def test_shell_type(self, shell):
        assert shell.type == "shell"

    def test_back_pointer_to_lump(self, shell):
        assert shell.lump.shell is shell

    def test_has_no_next_shell(self, shell):
        assert shell.next_shell.is_none is True

    def test_get_all_faces_as_a_list(self, shell):
        faces = shell.faces()
        assert len(faces) == 6
        assert isinstance(faces[0], entities.Face)

    def test_append_faces(self):
        body = entities.Body()
        lump = entities.Lump()
        shell = entities.Shell()
        body.append_lump(lump)
        lump.append_shell(shell)

        assert len(shell.faces()) == 0
        shell.append_face(entities.Face())
        assert len(shell.faces()) == 1
        shell.append_face(entities.Face())
        assert len(shell.faces()) == 2


class TestFace:
    @pytest.fixture(scope="class")
    def face(self, body):
        return body.lump.shell.face

    def test_face_type(self, face):
        assert face.type == "face"

    def test_back_pointer_to_shell(self, body, face):
        assert face.shell is body.lump.shell

    def test_has_attribute_surface(self, face):
        assert face.surface.type == "plane-surface"

    def test_face_features(self, face):
        assert face.sense is False  # forward
        assert face.double_sided is False  # single
        assert face.containment is False

    def test_traverse_all_six_cube_faces(self, face):
        count = 1
        while not face.next_face.is_none:
            count += 1
            face = face.next_face
        assert count == 6


class TestPolyhedronFaceBuilder:
    @pytest.fixture(scope="class")
    def cube(self):
        from ezdxf.render.forms import cube
        return mesh.PolyhedronFaceBuilder(cube())

    def test_creates_six_faces(self, cube):
        faces = list(cube.acis_faces())
        assert len(faces) == 6

    def test_each_face_defines_a_plane_surface(self, cube):
        for face in cube.acis_faces():
            assert face.surface.type == "plane-surface"


class TestPlane:
    @pytest.fixture(scope="class")
    def plane(self, body):
        return body.lump.shell.face.surface

    def test_plane_type(self, plane):
        assert plane.type == "plane-surface"

    def test_plane_location(self, plane):
        assert plane.origin.isclose((0, 0, 388.5))

    def test_plane_normal(self, plane):
        assert plane.normal.isclose((0, 0, 1))

    def test_plane_u_dir(self, plane):
        assert plane.u_dir.isclose((1, 0, 0))

    def test_plane_has_infinite_bounds(self, plane):
        assert math.isinf(plane.u_bounds[0])
        assert math.isinf(plane.u_bounds[1])
        assert math.isinf(plane.v_bounds[0])
        assert math.isinf(plane.v_bounds[1])


class TestLoop:
    @pytest.fixture(scope="class")
    def loop(self, body):
        return body.lump.shell.face.loop

    def test_loop_type(self, loop):
        assert loop.type == "loop"

    def test_cube_face_has_only_one_loop(self, loop):
        assert loop.next_loop.is_none is True

    def test_loop_references_the_parent_face(self, loop):
        assert loop.face.loop is loop

    def test_set_closed_coedges(self):
        loop = entities.Loop()
        coedges = [entities.Coedge() for _ in range(3)]
        loop.set_coedges(coedges, close=True)

        result = loop.coedges()
        assert result == coedges

    def test_set_open_coedges(self):
        loop = entities.Loop()
        coedges = [entities.Coedge() for _ in range(3)]
        loop.set_coedges(coedges, close=False)
        assert loop.coedge.prev_coedge.is_none is True

        result = loop.coedges()
        assert result == coedges


class TestCoedge:
    @pytest.fixture(scope="class")
    def coedge(self, body):
        return body.lump.shell.face.loop.coedge

    def test_co_edge_type(self, coedge):
        assert coedge.type == "coedge"

    def test_co_edges_are_organized_as_a_forward_linked_list(self, coedge):
        next = coedge.next_coedge
        co_edges = [next]
        while next is not coedge:
            next = next.next_coedge
            co_edges.append(next)
        assert len(co_edges) == 4

    def test_co_edges_are_organized_as_a_reverse_linked_list(self, coedge):
        prev = coedge.prev_coedge
        co_edges = [prev]
        while prev is not coedge:
            prev = prev.prev_coedge
            co_edges.append(prev)
        assert len(co_edges) == 4

    def test_partner_co_edge_count_is_two(self, coedge):
        assert len(coedge.partner_coedges()) == 1

    def test_co_edges_have_partner_co_edges_other_faces(self, coedge):
        assert coedge.partner_coedge.partner_coedge is coedge

    def test_sense_of_co_edge_is_forward(self, coedge):
        assert coedge.sense is False

    def test_co_edge_references_the_parent_loop(self, coedge):
        assert coedge.loop.coedge is coedge


def test_default_partner_co_edge_count_is_two():
    coedge = entities.Coedge()
    assert len(coedge.partner_coedges()) == 0


class TestEdge:
    @pytest.fixture(scope="class")
    def edge(self, body):
        return body.lump.shell.face.loop.coedge.edge

    def test_edge_type(self, edge):
        assert edge.type == "edge"

    def test_edge_has_a_start_vertex(self, edge):
        assert edge.start_vertex.is_none is False

    def test_edge_has_an_end_vertex(self, edge):
        assert edge.end_vertex.is_none is False

    # start- and end parameter do not exist in ACIS-400

    def test_sense_of_edge_is_forward(self, edge):
        assert edge.sense is False

    def test_underlying_curve_of_edge(self, edge):
        assert edge.curve.type == "straight-curve"

    def test_edge_is_referenced_by_two_parent_co_edges(self, edge):
        parent = edge.coedge
        assert parent.edge is edge
        assert parent.partner_coedge.edge is edge


class TestVertex:
    @pytest.fixture(scope="class")
    def vertex(self, body):
        return body.lump.shell.face.loop.coedge.edge.start_vertex

    def test_vertex_type(self, vertex):
        assert vertex.type == "vertex"

    def test_vertex_references_parent_edge(self, vertex):
        assert vertex.edge.start_vertex is vertex


class TestPoint:
    @pytest.fixture(scope="class")
    def point(self, body):
        return body.lump.shell.face.loop.coedge.edge.start_vertex.point

    def test_point_type(self, point):
        assert point.type == "point"

    def test_point_location(self, point):
        assert point.location.isclose((388.5, -388.5, 388.5))

    def test_get_all_points(self, body):
        face = body.lump.shell.face
        vertices = set()
        while not face.is_none:
            first_coedge = face.loop.coedge
            coedge = first_coedge
            while True:
                vertices.add(coedge.edge.start_vertex.point.location)
                coedge = coedge.next_coedge
                if coedge is first_coedge:
                    break
            face = face.next_face
        assert len(vertices) == 8


@pytest.fixture(scope="module")
def prism700(prism_sat):
    return load(prism_sat)


class TestExportSat:
    def test_export_rejects_unsupported_acis_versions(self, prism700):
        with pytest.raises(ExportError):
            export_sat(prism700, version=400)

    def test_export_acis_700(self, prism700):
        data = export_sat(prism700)
        assert len(data) == 117  # includes header
        assert data[-1] == "End-of-ACIS-data "  # an extra space at the end!


class TestExportSab21800:
    def test_export_rejects_unsupported_acis_versions(self, prism700):
        with pytest.raises(ExportError):
            export_sab(prism700, version=400)

    def test_reload_records_from_acis_export(self, prism700):
        data = export_sab(prism700)
        decoder = sab.Decoder(data)
        header = decoder.read_header()
        assert header.version == 21800
        records = list(decoder.read_records())
        # 115 = excludes header
        assert len(records) == 115
        assert records[-1][0].value == "End-of-ASM-data"


def test_load_mesh_from_exported_sat_data(prism700):
    bodies = load(export_sat(prism700))
    m = mesh.mesh_from_body(bodies[0])[0]
    assert len(m.vertices) == 8
    assert len(m.faces) == 10


def test_load_mesh_from_exported_sab_data(prism700):
    bodies = load(export_sab(prism700))
    m = mesh.mesh_from_body(bodies[0])[0]
    assert len(m.vertices) == 8
    assert len(m.faces) == 10


class TestExportTransform:
    @pytest.fixture(scope="class")
    def header(self):
        header = hdr.AcisHeader()
        header.version = 700
        return header

    @pytest.fixture(scope="class")
    def sat_exporter(self, header):
        return sat.SatExporter(header)

    @pytest.fixture(scope="class")
    def sab_exporter(self, header):
        return sab.SabExporter(header)

    def test_export_sat_identity_matrix(self, sat_exporter):
        data = []
        exporter = sat.SatDataExporter(sat_exporter, data)
        t = entities.Transform()
        t.matrix = Matrix44()
        t.export(exporter)
        assert (
            " ".join(data)
            == "1 0 0 0 1 0 0 0 1 0 0 0 1 no_rotate no_reflect no_shear"
        )

    def test_export_sab_identity_matrix(self, sab_exporter):
        data = []
        exporter = sab.SabDataExporter(sab_exporter, data)
        t = entities.Transform()
        t.matrix = Matrix44()
        t.export(exporter)
        assert data[0] == (
            const.Tags.LITERAL_STR,
            # The last space is important!
            "1 0 0 0 1 0 0 0 1 0 0 0 1 no_rotate no_reflect no_shear ",
        )


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2020 Manfred Moitzi
# License: MIT License
import pytest
from ezdxf.math import Vec2, distance_point_line_2d, is_point_on_line_2d


def test_distance_point_horiz_line():
    start, end = Vec2(0, 0), Vec2(10, 0)
    assert distance_point_line_2d(Vec2(5, 2), start, end) == 2
    assert distance_point_line_2d(Vec2(5, -3), start, end) == 3
    assert distance_point_line_2d(Vec2(0, 0), start, end) == 0


def test_distance_point_vertical_line():
    start, end = Vec2(0, 0), Vec2(0, 10)
    assert distance_point_line_2d(Vec2(2, 0), start, end) == 2
    assert distance_point_line_2d(Vec2(-3, 0), start, end) == 3
    assert distance_point_line_2d(Vec2(0, 0), start, end) == 0


def test_is_point_on_horiz_line():
    assert is_point_on_line_2d(Vec2(-1, 0), Vec2(0, 0), Vec2(1, 0)) is True
    assert (
        is_point_on_line_2d(Vec2(-1, 0), Vec2(0, 0), Vec2(1, 0.0001)) is False
    )
    assert (
        is_point_on_line_2d(Vec2(-1, 0), Vec2(0, 0), Vec2(1, 0), ray=False)
        is False
    )


def test_is_point_on_vertical_line():
    assert is_point_on_line_2d(Vec2(0, -1), Vec2(0, 0), Vec2(0, 1)) is True
    assert (
        is_point_on_line_2d(Vec2(0, -1), Vec2(0, 0), Vec2(0.00001, 1)) is False
    )
    assert (
        is_point_on_line_2d(Vec2(0, -1), Vec2(0, 0), Vec2(0, 1), ray=False)
        is False
    )


def test_is_point_on_diag_line():
    assert is_point_on_line_2d(Vec2(-1, -1), Vec2(0, 0), Vec2(1, 1)) is True
    assert (
        is_point_on_line_2d(Vec2(-1, -1), Vec2(0, 0), Vec2(1, 1), ray=False)
        is False
    )


if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License

import pytest
from ezdxf.entities import (
    DXFObject,
    is_dxf_object,
    DXFTagStorage,
    is_graphic_entity,
)

XRECORD = """  0
XRECORD
5
0
330
0
100
AcDbXrecord
280
1
"""


@pytest.fixture
def entity():
    return DXFTagStorage.from_text(XRECORD)


def test_is_dxf_object():
    assert is_dxf_object(DXFObject()) is True


def test_wrapped_xrecord_is_a_dxf_object(entity):
    assert is_dxf_object(entity) is True


def test_wrapped_xrecord_is_not_a_graphic_entity(entity):
    assert is_graphic_entity(entity) is False


if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2022-2023, Manfred Moitzi
#  License: MIT License

"""
This module provides support for fonts stored as SHX and SHP files.
(SHP is not the GIS file format!)

The documentation about the SHP file format can be found at Autodesk:
https://help.autodesk.com/view/OARX/2018/ENU/?guid=GUID-DE941DB5-7044-433C-AA68-2A9AE98A5713

Using bytes for strings because no encoding is defined in shape-files.

"""
from __future__ import annotations
from typing import (
    Sequence,
    Iterable,
    Iterator,
    Callable,
    Optional,
    Any,
    Set,
    no_type_check,
)
import dataclasses
import enum
import math
import pathlib

from ezdxf.math import (
    UVec,
    Vec2,
    ConstructionEllipse,
    bulge_to_arc,
    BoundingBox2d,
    Matrix44,
)
from ezdxf import path
from .font_measurements import FontMeasurements
from .glyphs import GlyphPath, Glyphs


class ShapeFileException(Exception):
    pass


class UnsupportedShapeFile(ShapeFileException):
    pass


class UnsupportedShapeNumber(ShapeFileException):
    pass


class InvalidFontDefinition(ShapeFileException):
    pass


class InvalidFontParameters(ShapeFileException):
    pass


class InvalidShapeRecord(ShapeFileException):
    pass


class FileStructureError(ShapeFileException):
    pass


class StackUnderflow(ShapeFileException):
    pass


class FontEmbedding(enum.IntEnum):
    ALLOWED = 0
    DISALLOWED = 1
    READONLY = 2


class FontEncoding(enum.IntEnum):
    UNICODE = 0
    PACKED_MULTIBYTE_1 = 1
    SHAPE_FILE = 2


class FontMode(enum.IntEnum):
    HORIZONTAL = 0
    BIDIRECT = 2


NO_DATA: Sequence[int] = tuple()
DEBUG = False
DEBUG_CODES: Set[int] = set()
DEBUG_SHAPE_NUMBERS = set()
ORD_NULL = ord("0")
ORD_MINUS = ord("-")


# slots=True - Python 3.10+
@dataclasses.dataclass
class Symbol:
    number: int
    byte_count: int
    name: bytes
    data: Sequence[int] = NO_DATA

    def export_str(self, as_num: int = 0) -> list[bytes]:
        num = as_num if as_num else self.number
        export = [f"*{num:05X},{self.byte_count},".encode() + self.name]
        export.extend(format_shape_data_string(self.data))
        return export


def format_shape_data_string(data: Sequence[int]) -> list[bytes]:
    export = []
    s = b""
    for num in data:
        s2 = s + f"{num},".encode()
        if len(s2) < 80:
            s = s2
        else:
            export.append(s[:-1])
            s = f",{num},".encode()
    if s:
        export.append(s[:-1])
    return export


class ShapeFile:
    """Low level representation of a SHX/SHP file."""
    def __init__(
        self,
        name: bytes,
        above: int,
        below: int,
        mode=FontMode.HORIZONTAL,
        encoding=FontEncoding.UNICODE,
        embed=FontEmbedding.ALLOWED,
    ):
        self.shapes: dict[int, Symbol] = dict()
        self.name = name
        self.above = above
        self.below = below
        self.mode = mode
        self.encoding = encoding
        self.embed = embed

    @property
    def cap_height(self) -> float:
        return float(self.above)

    @property
    def descender(self) -> float:
        return float(self.below)

    @property
    def is_font(self) -> bool:
        return self.encoding != FontEncoding.SHAPE_FILE

    @property
    def is_shape_file(self) -> bool:
        return self.encoding == FontEncoding.SHAPE_FILE

    def find(self, name: bytes) -> Optional[Symbol]:
        for symbol in self.shapes.values():
            if name == symbol.name:
                return symbol
        return None

    def __len__(self):
        return len(self.shapes)

    def __getitem__(self, item):
        return self.shapes[item]

    @staticmethod
    def from_ascii_records(records: Sequence[bytes]):
        if len(records) == 2:
            encoding = FontEncoding.UNICODE
            above = 0
            below = 0
            embed = FontEmbedding.ALLOWED
            mode = FontMode.HORIZONTAL
            end = 0
            header, params = records
            try:
                # ignore second value: defbytes
                spec, _, name = header.split(b",", maxsplit=2)
            except ValueError:
                raise InvalidFontDefinition()
            if spec == b"*UNIFONT":
                try:
                    above, below, mode, encoding, embed, end = params.split(b",")  # type: ignore
                except ValueError:
                    raise InvalidFontParameters(params)
            elif spec == b"*0":
                try:
                    above, below, mode, *rest = params.split(b",")  # type: ignore
                    end = rest[-1]  # type: ignore
                except ValueError:
                    raise InvalidFontParameters(params)
            else:  # it's a simple shape file
                encoding = FontEncoding.SHAPE_FILE
            assert int(end) == 0

            return ShapeFile(
                name.strip(),
                int(above),
                int(below),
                FontMode(int(mode)),
                FontEncoding(int(encoding)),
                FontEmbedding(int(embed)),
            )

    def parse_ascii_records(self, records: Iterable[Sequence[bytes]]) -> None:
        for record in records:
            if len(record) < 2:
                raise InvalidShapeRecord(record)
            try:
                number, byte_count, name = split_def_record(record[0])
            except ValueError:
                raise FileStructureError(record[0])

            assert len(number) > 1
            if number[1] == ORD_NULL:  # hex if first char is "0" like "*0000A"
                int_num = int(number[1:], 16)
            else:  # like "*130"
                int_num = int(number[1:], 10)

            symbol = Symbol(int_num, int(byte_count), name)
            data = b"".join(record[1:])
            symbol.data = tuple(parse_codes(split_record(data)))
            if symbol.data[-1] == 0:
                self.shapes[int_num] = symbol
            else:
                s = record[0].decode(errors="ignore")
                raise FileStructureError(f"file structure error at symbol <{s}>")

    def get_codes(self, number: int) -> Sequence[int]:
        symbol = self.shapes.get(number)
        if symbol is None:
            raise UnsupportedShapeNumber(number)
        return symbol.data

    def render_shape(self, number: int, stacked=False) -> GlyphPath:
        return render_shapes([number], self.get_codes, stacked=stacked)

    def render_shapes(self, numbers: Sequence[int], stacked=False) -> GlyphPath:
        return render_shapes(numbers, self.get_codes, stacked=stacked)

    def render_text(self, text: str, stacked=False) -> GlyphPath:
        numbers = [ord(char) for char in text]
        return render_shapes(
            numbers, self.get_codes, stacked=stacked, reset_to_baseline=True
        )

    def shape_string(self, shape_number: int, as_num: int = 0) -> list[bytes]:
        return self.shapes[shape_number].export_str(as_num)


def readfile(filename: str) -> ShapeFile:
    """Load shape-file `filename`, the file type is detected by the file
    name extension and should be ".shp" or ".shx" otherwise rises an
    exception :class:`UnsupportedShapeFile`.
    """
    data = pathlib.Path(filename).read_bytes()
    filename = filename.lower()
    if filename.endswith(".shp"):
        return shp_load(data)
    elif filename.endswith(".shx"):
        return shx_load(data)
    else:
        raise UnsupportedShapeFile("unknown filetype")


def split_def_record(record: bytes) -> Sequence[bytes]:
    return tuple(s.strip() for s in record.split(b",", maxsplit=2))


def split_record(record: bytes) -> Sequence[bytes]:
    return tuple(s.strip() for s in record.split(b","))


def parse_codes(codes: Iterable[bytes]) -> Iterator[int]:
    for code in codes:
        code = code.strip(b"()")
        if code == b"":
            continue
        if code[0] == ORD_NULL:
            yield int(code, 16)
        elif code[0] == ORD_MINUS and code[1] == ORD_NULL:
            yield int(code, 16)
        else:
            yield int(code, 10)


def shp_load(data: bytes) -> ShapeFile:
    records = parse_string_records(merge_lines(filter_noise(data.split(b"\n"))))
    if b"*UNIFONT" in records:
        font_definition = records.pop(b"*UNIFONT")
    elif b"*0" in records:
        font_definition = records.pop(b"*0")
    else:
        # a common shape file without a name
        # symbol numbers are decimal!!!
        font_definition = (b"_,_,_", b"")
    shp = ShapeFile.from_ascii_records(font_definition)
    shp.parse_ascii_records(records.values())
    return shp


def shx_load(data: bytes) -> ShapeFile:
    if data.startswith(b"AutoCAD-86 shapes 1.0"):
        return load_shx_shape_file_1_0(data)
    elif data.startswith(b"AutoCAD-86 shapes 1.1"):
        return load_shx_shape_file_1_1(data)
    elif data.startswith(b"AutoCAD-86 unifont 1.0"):
        return load_shx_unifont_file_1_0(data)
    elif data.startswith(b"AutoCAD-86 bigfont 1.0"):
        raise UnsupportedShapeFile("BIGFONT shapes are not supported yet")
    raise UnsupportedShapeFile("unknown shape file format")


def load_shx_shape_file_1_0(data: bytes) -> ShapeFile:
    above = 0
    below = 0
    mode = FontMode.HORIZONTAL
    name = b""
    shapes = parse_shx_shapes(data)
    font_definition = shapes.get(0)
    if font_definition:
        shapes.pop(0)
        shape_data = font_definition.data
        above = shape_data[0]
        below = shape_data[1]
        mode = FontMode(shape_data[2])

    shape_file = ShapeFile(name, above=above, below=below, mode=mode)
    shape_file.shapes = shapes
    return shape_file


def load_shx_shape_file_1_1(data: bytes) -> ShapeFile:
    # seems to be the same as v1.0:
    return load_shx_shape_file_1_0(data)


class DataReader:
    def __init__(self, data: bytes, index=0):
        self.data = data
        self.index = index

    @property
    def has_data(self) -> bool:
        return self.index < len(self.data)

    def skip(self, n: int) -> None:
        self.index += n

    def u8(self) -> int:
        index = self.index
        self.index += 1
        return self.data[index]

    def i8(self) -> int:
        value = self.data[self.index]
        self.index += 1
        if value > 127:
            return (value & 127) - 128  # ???
        return value

    def octant(self) -> int:
        value = self.data[self.index]
        self.index += 1
        if value & 128:
            return -(value & 127)
        else:
            return value

    def read_str(self) -> bytes:
        data = bytearray()
        while True:
            char = self.u8()
            if char:
                data.append(char)
            else:
                return data

    def read_bytes(self, n: int) -> bytes:
        index = self.index
        self.index += n
        return self.data[index : index + n]

    def u16(self) -> int:
        # little ending
        index = self.index
        self.index += 2
        return self.data[index] + (self.data[index + 1] << 8)


# the old 'AutoCAD-86 shapes 1.0' format
SHX_SHAPES_START_INDEX = 0x17


def parse_shx_shapes(data: bytes) -> dict[int, Symbol]:
    """SHX/SHP shape parser."""
    shapes: dict[int, Symbol] = dict()
    reader = DataReader(data, SHX_SHAPES_START_INDEX)
    if reader.u8() != 0x1A:
        raise FileStructureError("signature byte 0x1A not found")
    first_number = reader.u16()
    last_number = reader.u16()
    shape_count = reader.u16()
    index_table: list[tuple[int, int]] = []
    for _ in range(shape_count):
        shape_number = reader.u16()
        data_size = reader.u16()
        index_table.append((shape_number, data_size))
    if index_table[0][0] != first_number:
        raise FileStructureError("invalid first entry in index table")
    if index_table[-1][0] != last_number:
        raise FileStructureError("invalid last entry in index table")
    for shape_number, length in index_table:
        record = reader.read_bytes(length)
        try:
            name, shape_data = parse_shx_data_record(record)
        except IndexError:
            raise FileStructureError(
                f"SHX parsing error shape *{shape_number:05X}: {str(record)}"
            )
        byte_count = length - len(name) - 1
        shapes[shape_number] = Symbol(shape_number, byte_count, name, shape_data)
    if reader.read_bytes(3) != b"EOF":
        raise FileStructureError("EOF marker not found")
    return shapes


def parse_shx_data_record(data: bytes) -> tuple[bytes, Sequence[int]]:
    """Low level SHX/SHP shape record parser."""
    reader = DataReader(data)
    name = reader.read_str()
    codes = parse_shape_codes(reader)
    return name, codes


# the new 'AutoCAD-86 unifont 1.0' format
SHX_UNIFONT_START_INDEX = 0x18


def load_shx_unifont_file_1_0(data: bytes) -> ShapeFile:
    reader = DataReader(data, SHX_UNIFONT_START_INDEX)
    name, above, below, mode, encoding, embed = parse_shx_unifont_definition(reader)
    shape_file = ShapeFile(
        name,
        above=above,
        below=below,
        mode=mode,
        encoding=encoding,
        embed=embed,
    )
    try:
        shape_file.shapes = parse_shx_unifont_shapes(reader)
    except IndexError:
        raise FileStructureError("pre-mature end of file")
    return shape_file


def parse_shx_unifont_definition(reader: DataReader) -> Sequence[Any]:
    if reader.u8() != 0x1A:
        raise FileStructureError("signature byte 0x1A not found")
    reader.skip(6)
    # u16 record count: in isocp.shx = 238 (0xEE 0x00)
    #     = 1 definition record + 237 character records
    # u16 isocp.shx: 0 (0x00 0x00), always 0???
    # u16 font definition record size; isocp.shx: 52 (0x34 0x00)

    name = reader.read_str()
    above = reader.u8()
    below = reader.u8()
    mode = FontMode(reader.u8())
    encoding = FontEncoding(reader.u8())
    embed = FontEmbedding(reader.u8())
    reader.skip(1)  # <00> byte - end of font definition
    return name, above, below, mode, encoding, embed


def parse_shx_unifont_shapes(reader: DataReader) -> dict[int, Symbol]:
    shapes: dict[int, Symbol] = dict()
    while reader.has_data:
        shape_number = reader.u16()
        byte_count = reader.u16()
        name = reader.read_str()
        byte_count = byte_count - len(name) - 1
        data_record = reader.read_bytes(byte_count)
        codes = parse_shape_codes(DataReader(data_record), unifont=True)
        shapes[shape_number] = Symbol(shape_number, byte_count, name, codes)
    return shapes


SINGLE_CODES = {1, 2, 5, 6, 14}


def parse_shape_codes(reader: DataReader, unifont=False) -> Sequence[int]:
    codes: list[int] = []
    while True:
        code = reader.u8()
        codes.append(code)
        if code == 0:
            return tuple(codes)
        elif code in SINGLE_CODES or code > 14:
            continue
        elif code == 3 or code == 4:  # size control
            codes.append(reader.u8())
        elif code == 7:  # sub shape
            if unifont:
                codes.append(reader.u16())
            else:
                codes.append(reader.u8())
        elif code == 8:  # displacement
            codes.append(reader.i8())
            codes.append(reader.i8())
        elif code == 9:  # multiple displacement
            x, y = 1, 1
            while x or y:
                x = reader.i8()
                y = reader.i8()
                codes.append(x)
                codes.append(y)
        elif code == 10:  # octant arc
            codes.append(reader.u8())  # radius
            codes.append(reader.octant())  # octant specs
        elif code == 11:  # fractional arc
            codes.append(reader.u8())  # start offset
            codes.append(reader.u8())  # end offset
            codes.append(reader.u8())  # radius hi
            codes.append(reader.u8())  # radius lo
            codes.append(reader.octant())  # octant specs
        elif code == 12:  # bulge arcs
            codes.append(reader.i8())  # x
            codes.append(reader.i8())  # y
            codes.append(reader.i8())  # bulge
        elif code == 13:  # multiple bulge arcs
            x, y = 1, 1
            while x or y:
                x = reader.i8()
                y = reader.i8()
                codes.append(x)
                codes.append(y)
                if x or y:
                    codes.append(reader.i8())


def filter_noise(lines: Iterable[bytes]) -> Iterator[bytes]:
    for line in lines:
        line.rstrip(b"\r\n")
        line = line.strip()
        if line:
            line = line.split(b";")[0]
            line = line.strip()
            if line:
                yield line


def merge_lines(lines: Iterable[bytes]) -> Iterator[bytes]:
    current = b""
    for next_line in lines:
        if not current:
            current = next_line
        elif current.startswith(b"*"):
            if next_line.startswith(b","):  # wrapped specification line?
                current += next_line
            else:
                yield current
                current = next_line
        elif current.endswith(b","):
            current += next_line
        elif next_line.startswith(b","):
            current += next_line
        else:
            yield current
            current = next_line
    if current:
        yield current


def parse_string_records(
    lines: Iterable[bytes],
) -> dict[bytes, Sequence[bytes]]:
    records: dict[bytes, Sequence[bytes]] = dict()
    name = None
    record = []
    for line in lines:
        if line.startswith(b"*BIGFONT"):
            raise UnsupportedShapeFile("BIGFONT shape files are not supported yet")
        if line.startswith(b"*"):
            if name is not None:
                records[name] = tuple(record)
            name = line.split(b",")[0].strip()
            record = [line]
        else:
            record.append(line)

    if name is not None:
        records[name] = tuple(record)
    return records


def render_shapes(
    shape_numbers: Sequence[int],
    get_codes: Callable[[int], Sequence[int]],
    stacked: bool,
    start: UVec = (0, 0),
    reset_to_baseline=False,
) -> GlyphPath:
    """Renders multiple shapes into a single glyph path."""
    ctx = ShapeRenderer(
        path.Path(start),
        pen_down=True,
        stacked=stacked,
        get_codes=get_codes,
    )
    for shape_number in shape_numbers:
        try:
            ctx.render(shape_number, reset_to_baseline=reset_to_baseline)
        except UnsupportedShapeNumber:
            pass
        except StackUnderflow:
            raise StackUnderflow(
                f"stack underflow while rendering shape number {shape_number}"
            )
        # move cursor to the start of the next char???
    return GlyphPath(ctx.p)


#        0, 1, 2,   3, 4,    5,  6,  7,  8,  9,  A,    B, C,   D, E, F
VEC_X = [1, 1, 1, 0.5, 0, -0.5, -1, -1, -1, -1, -1, -0.5, 0, 0.5, 1, 1]
#        0,   1, 2, 3, 4, 5, 6,   7, 8,    9,  A,  B,  C,  D,  E,    F
VEC_Y = [0, 0.5, 1, 1, 1, 1, 1, 0.5, 0, -0.5, -1, -1, -1, -1, -1, -0.5]


class ShapeRenderer:
    """Low level glyph renderer for SHX/SHP fonts."""
    def __init__(
        self,
        p: path.Path,
        get_codes: Callable[[int], Sequence[int]],
        *,
        vector_length: float = 1.0,
        pen_down: bool = True,
        stacked: bool = False,
    ):
        self.p = p
        self.vector_length = float(vector_length)  # initial vector length
        self.pen_down = pen_down
        self.stacked = stacked  # vertical stacked text
        self._location_stack: list[Vec2] = []
        self._get_codes = get_codes
        self._baseline_y = self.p.start.y

    @property
    def current_location(self) -> Vec2:
        return Vec2(self.p.end)

    def push(self) -> None:
        self._location_stack.append(self.current_location)

    def pop(self) -> None:
        self.p.move_to(self._location_stack.pop())

    def render(
        self,
        shape_number: int,
        reset_to_baseline=False,
    ) -> None:
        self.pen_down = True
        codes = self._get_codes(shape_number)
        index = 0
        skip_next = False
        while index < len(codes):
            code = codes[index]
            if DEBUG and (code in DEBUG_CODES) and shape_number:
                DEBUG_SHAPE_NUMBERS.add(shape_number)
            index += 1
            if code > 15 and not skip_next:
                self.draw_vector(code)
            elif code == 0:
                break
            elif code == 1 and not skip_next:  # pen down
                self.pen_down = True
            elif code == 2 and not skip_next:  # pen up
                self.pen_down = False
            elif code == 3 or code == 4:  # scale size
                factor = codes[index]
                index += 1
                if not skip_next:
                    if code == 3:
                        self.vector_length /= factor
                    elif code == 4:
                        self.vector_length *= factor
                    continue
            elif code == 5 and not skip_next:  # push location state
                self.push()
            elif code == 6 and not skip_next:  # pop location state
                try:
                    self.pop()
                except IndexError:
                    raise StackUnderflow()
            elif code == 7:  # sub-shape:
                sub_shape_number = codes[index]
                index += 1
                if not skip_next:
                    # Use current state of pen and location!
                    self.render(sub_shape_number)
                    # resume with current state of pen and location!
            elif code == 8:  # displacement vector
                x = codes[index]
                y = codes[index + 1]
                index += 2
                if not skip_next:
                    self.draw_displacement(x, y)
            elif code == 9:  # multiple displacements vectors
                while True:
                    x = codes[index]
                    y = codes[index + 1]
                    index += 2
                    if x == 0 and y == 0:
                        break
                    if not skip_next:
                        self.draw_displacement(x, y)
            elif code == 10:  # 10 octant arc
                radius = codes[index]
                start_octant, octant_span, ccw = decode_octant_specs(codes[index + 1])
                if octant_span == 0:  # full circle
                    octant_span = 8
                index += 2
                if not skip_next:
                    self.draw_arc_span(
                        radius * self.vector_length,
                        math.radians(start_octant * 45),
                        math.radians(octant_span * 45),
                        ccw,
                    )
            elif code == 11:  # fractional arc
                # TODO: this still seems not 100% correct, see vertical placing
                #  of characters "9" and "&" for font isocp.shx.
                #  This is solved by placing the end point on the baseline after
                #  each character rendering, but only for text rendering.
                #  The remaining problems can possibly be due to a loss of
                #  precision when converting floats to ints.
                start_offset = codes[index]
                end_offset = codes[index + 1]
                radius = (codes[index + 2] << 8) + codes[index + 3]
                start_octant, octant_span, ccw = decode_octant_specs(codes[index + 4])
                index += 5
                if end_offset == 0:
                    end_offset = 256
                binary_deg = 45.0 / 256.0
                start_offset_angle = start_offset * binary_deg
                end_offset_angle = end_offset * binary_deg
                if ccw:
                    end_octant = start_octant + octant_span - 1
                    start_angle = start_octant * 45.0 + start_offset_angle
                    end_angle = end_octant * 45.0 + end_offset_angle
                else:
                    end_octant = start_octant - octant_span + 1
                    start_angle = start_octant * 45.0 - start_offset_angle
                    end_angle = end_octant * 45.0 - end_offset_angle

                if not skip_next:
                    self.draw_arc_start_to_end(
                        radius * self.vector_length,
                        math.radians(start_angle),
                        math.radians(end_angle),
                        ccw,
                    )
            elif code == 12:  # bulge arc
                x = codes[index]
                y = codes[index + 1]
                bulge = codes[index + 2]
                index += 3
                if not skip_next:
                    self.draw_bulge(x, y, bulge)
            elif code == 13:  # multiple bulge arcs
                while True:
                    x = codes[index]
                    y = codes[index + 1]
                    if x == 0 and y == 0:
                        index += 2
                        break
                    bulge = codes[index + 2]
                    index += 3
                    if not skip_next:
                        self.draw_bulge(x, y, bulge)
            elif code == 14:  # flag vertical text
                if not self.stacked:
                    skip_next = True
                    continue
            skip_next = False

        if reset_to_baseline:
            # HACK: because of invalid fractional arc rendering!
            if not math.isclose(self.p.end.y, self._baseline_y):
                self.p.move_to((self.p.end.x, self._baseline_y))

    def draw_vector(self, code: int) -> None:
        angle: int = code & 0xF
        length: int = (code >> 4) & 0xF
        self.draw_displacement(VEC_X[angle] * length, VEC_Y[angle] * length)

    def draw_displacement(self, x: float, y: float):
        scale = self.vector_length
        target = self.current_location + Vec2(x * scale, y * scale)
        if self.pen_down:
            self.p.line_to(target)
        else:
            self.p.move_to(target)

    def draw_arc_span(
        self, radius: float, start_angle: float, span_angle: float, ccw: bool
    ):
        # IMPORTANT: radius has to be scaled by self.vector_length!
        end_angle = start_angle + (span_angle if ccw else -span_angle)
        self.draw_arc_start_to_end(radius, start_angle, end_angle, ccw)

    def draw_arc_start_to_end(
        self, radius: float, start_angle: float, end_angle: float, ccw: bool
    ):
        # IMPORTANT: radius has to be scaled by self.vector_length!
        assert radius > 0.0
        arc = ConstructionEllipse(
            major_axis=(radius, 0),
            start_param=start_angle,
            end_param=end_angle,
            ccw=ccw,
        )
        # arc goes start -> end if ccw otherwise end -> start
        # move arc start-point to the end-point of current path
        arc.center += self.current_location - Vec2(
            arc.start_point if ccw else arc.end_point
        )
        if self.pen_down:
            path.add_ellipse(self.p, arc, reset=False)
        else:
            self.p.move_to(arc.end_point if ccw else arc.start_point)

    def draw_arc(
        self,
        center: Vec2,
        radius: float,
        start_param: float,
        end_param: float,
        ccw: bool,
    ):
        # IMPORTANT: radius has to be scaled by self.vector_length!
        arc = ConstructionEllipse(
            center=center,
            major_axis=(radius, 0),
            start_param=start_param,
            end_param=end_param,
            ccw=ccw,
        )
        if self.pen_down:
            path.add_ellipse(self.p, arc, reset=False)
        else:
            self.p.move_to(arc.end_point if ccw else arc.start_point)

    def draw_bulge(self, x: float, y: float, bulge: float):
        if self.pen_down and bulge:
            start_point = self.current_location
            scale = self.vector_length
            ccw = bulge > 0
            end_point = start_point + Vec2(x * scale, y * scale)
            bulge = abs(bulge) / 127.0
            if ccw:  # counter-clockwise
                center, start_angle, end_angle, radius = bulge_to_arc(
                    start_point, end_point, bulge
                )
            else:  # clockwise
                center, start_angle, end_angle, radius = bulge_to_arc(
                    end_point, start_point, bulge
                )
            self.draw_arc(center, radius, start_angle, end_angle, ccw=True)
        else:
            self.draw_displacement(x, y)


def decode_octant_specs(specs: int) -> tuple[int, int, bool]:
    ccw = True
    if specs < 0:
        ccw = False
        specs = -specs
    start_octant = (specs >> 4) & 0xF
    octant_span = specs & 0xF
    return start_octant, octant_span, ccw


def find_shape_number(filename: str, name: bytes) -> int:
    """Returns the shape number for shape `name` from the shape-file, returns -1 if not
    found.

    The filename can be an absolute path or the shape-file will be searched in the
    current directory and all directories stored in :attr:`ezdxf.options.support_dirs`.
    Supports shape files of type ".shx" and ".shp".
    """
    from ezdxf.filemanagement import find_support_file
    from ezdxf import options

    shape_file_name = find_support_file(filename, options.support_dirs)
    index = -1
    try:
        shapes = readfile(shape_file_name)
    except (IOError, ShapeFileException):
        return index
    shx_symbol = shapes.find(name)
    if shx_symbol is not None:
        return shx_symbol.number
    return index


class GlyphCache(Glyphs):
    def __init__(self, font: ShapeFile) -> None:
        """Text render engine for SHX/SHP fonts with integrated glyph caching."""
        self.font: ShapeFile = font
        self._glyph_cache: dict[int, GlyphPath] = dict()
        self._advance_width_cache: dict[int, float] = dict()
        self.space_width: float = self.detect_space_width(default_width=font.cap_height)
        self.empty_box: GlyphPath = self.get_empty_box()
        self.font_measurements: FontMeasurements = self._get_font_measurements()

    def get_scaling_factor(self, cap_height: float) -> float:
        try:
            return cap_height / self.font_measurements.cap_height
        except ZeroDivisionError:
            return 1.0

    def detect_space_width(self, default_width: float) -> float:
        # space (32) or "A" (65) may not present!
        if 32 in self.font.shapes:
            return self.get_shape(32).end.x  # width of space
        if 65 in self.font.shapes:
            return self.get_shape(65).end.x  # width of "A"
        return default_width

    def get_empty_box(self, char: int = 65) -> GlyphPath:
        """Returns the empty box (tofu) representation based on the size of the given
        character index.

        The given index may not exist and a default box based on cap-height and space
        width will be created.
        """
        # default box parameters
        font = self.font
        height = font.cap_height
        width = self.space_width
        start = Vec2(0, 0)
        end = Vec2(width, 0)

        if char in font.shapes:
            # box parameters based on the glyph size
            glyph = self.get_shape(char)
            box = BoundingBox2d(glyph.control_vertices())
            height = box.size.y
            width = box.size.x
            start = glyph.start
            end = glyph.end

        p = path.Path(start)
        p.line_to(start + Vec2(width, 0))
        p.line_to(start + Vec2(width, height))
        p.line_to(start + Vec2(0, height))
        p.close()
        p.move_to(end)
        return GlyphPath(p)

    def _render_shape(self, shape_number) -> GlyphPath:
        ctx = ShapeRenderer(
            path.Path(),
            pen_down=True,
            stacked=False,
            get_codes=self.font.get_codes,
        )
        try:
            ctx.render(shape_number, reset_to_baseline=False)
        except StackUnderflow:
            pass
        return GlyphPath(ctx.p)

    def get_shape(self, shape_number: int) -> GlyphPath:
        try:
            return self._glyph_cache[shape_number].clone()
        except KeyError:
            pass
        try:
            glyph = self._render_shape(shape_number)
        except UnsupportedShapeNumber:
            if shape_number < 32:
                glyph = GlyphPath(None)
            else:
                glyph = self.empty_box
        self._glyph_cache[shape_number] = glyph
        try:
            width = glyph.end.x
        except IndexError:
            width = self.space_width
        self._advance_width_cache[shape_number] = width
        return glyph.clone()

    def get_advance_width(self, shape_number: int) -> float:
        if shape_number == 32:
            return self.space_width
        try:
            return self._advance_width_cache[shape_number]
        except KeyError:
            pass
        return self.get_shape(shape_number).end.x

    @no_type_check
    def _get_font_measurements(self) -> FontMeasurements:
        # ignore last move_to command, which places the pen at the start of the
        # following glyph
        bbox = BoundingBox2d(self.get_shape(ord("x")).control_vertices()[:-1])
        baseline = bbox.extmin.y
        x_height = bbox.extmax.y - baseline

        cap_height = self.font.above
        if cap_height == 0:
            bbox = BoundingBox2d(self.get_shape(ord("A")).control_vertices()[:-1])
            cap_height = bbox.extmax.y - baseline
        descender_height = self.font.below
        if descender_height == 0:
            bbox = BoundingBox2d(self.get_shape(ord("p")).control_vertices()[:-1])
            descender_height = baseline - bbox.extmin.y
        return FontMeasurements(
            baseline=baseline,
            cap_height=cap_height,
            x_height=x_height,
            descender_height=descender_height,
        )

    def get_text_length(
        self, text: str, cap_height: float, width_factor: float = 1.0
    ) -> float:
        scaling_factor = self.get_scaling_factor(cap_height) * width_factor
        return sum(self.get_advance_width(ord(c)) for c in text) * scaling_factor

    def get_text_glyph_paths(
        self, text: str, cap_height: float = 1.0, width_factor: float = 1.0
    ) -> list[GlyphPath]:
        """Returns the glyph paths of string `s` as a list, scaled to cap height."""
        glyph_paths: list[GlyphPath] = []
        sy = self.get_scaling_factor(cap_height)
        sx = sy * width_factor
        m = Matrix44.scale(sx, sy, 1)
        current_location = 0.0
        for c in text:
            shape_number = ord(c)
            if shape_number > 32:
                glyph = self.get_shape(shape_number)
                m[3, 0] = current_location
                glyph.transform_inplace(m)
                glyph_paths.append(glyph)
            current_location += self.get_advance_width(shape_number) * sx
        return glyph_paths



# Copyright (c) 2024, Manfred Moitzi
# License: MIT License
import pytest

from pathlib import Path
import ezdxf

DATA = Path(__file__).parent / "data"


def test_load_r12_blocks_with_no_name():
    """DXF R12 has no BLOCK_RECORD and therefore the block name cannot be recovered."""
    doc = ezdxf.readfile(DATA / "r12_blocks_with_no_names.dxf")
    assert len(doc.blocks) == 2


def test_load_r2000_blocks_with_recovered_names():
    """DXF R2000 requires a BLOCK_RECORD for each block definition and therefore the
    block name can be recovered from the BLOCK_RECORD.
    """
    doc = ezdxf.readfile(DATA / "r2000_blocks_with_no_names.dxf")
    assert len(doc.blocks) == 4
    for name in ("Test", "Test2"):
        block = doc.blocks.get(name)
        assert block.name == name
        assert block.block.dxf.name == name
        assert block.block_record.dxf.name == name


def test_load_r2000_blocks_with_no_name_and_no_block_record():
    """The block name cannot be recovered without an associated block record."""
    doc = ezdxf.readfile(DATA / "r2000_blocks_with_no_name_and_no_block_record.dxf")
    assert len(doc.blocks) == 2
    for name in ("Test", "Test2"):
        assert name not in doc.blocks
        assert name not in doc.block_records


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2016-2023, Manfred Moitzi
# License: MIT License
import pytest
import logging

import ezdxf
from ezdxf.entities.image import ImageDef, Image

logger = logging.getLogger("ezdxf")


@pytest.fixture(scope="module")
def doc():
    return ezdxf.new("R2000")


@pytest.fixture(scope="module")
def image_def(doc):
    return ImageDef.from_text(IMAGE_DEF, doc)


def test_set_raster_variables():
    doc = ezdxf.new("R2000")
    assert "ACAD_IMAGE_VARS" not in doc.rootdict
    doc.set_raster_variables(frame=0, quality=1, units="m")
    raster_vars = doc.rootdict["ACAD_IMAGE_VARS"]
    assert raster_vars.dxftype() == "RASTERVARIABLES"
    assert raster_vars.dxf.frame == 0
    assert raster_vars.dxf.quality == 1
    assert raster_vars.dxf.units == 3  # m

    assert "ACAD_IMAGE_VARS" in doc.rootdict
    doc.set_raster_variables(frame=1, quality=0, units="km")
    raster_vars = doc.rootdict["ACAD_IMAGE_VARS"]
    assert raster_vars.dxftype() == "RASTERVARIABLES"
    assert raster_vars.dxf.frame == 1
    assert raster_vars.dxf.quality == 0
    assert raster_vars.dxf.units == 4  # km


def test_imagedef_attribs(image_def):
    assert image_def.dxftype() == "IMAGEDEF"
    assert image_def.dxf.class_version == 0
    assert image_def.dxf.filename == "mycat.jpg"
    assert image_def.dxf.image_size.isclose((640.0, 360.0))
    assert image_def.dxf.pixel_size.isclose((0.01, 0.01))
    assert image_def.dxf.loaded == 1
    assert image_def.dxf.resolution_units == 0


@pytest.fixture(scope="module")
def image(doc):
    return Image.from_text(IMAGE, doc)


def test_image_dxf_attribs(image):
    assert "IMAGE" == image.dxftype()
    assert image.dxf.insert.isclose((0.0, 0.0, 0.0))
    assert image.dxf.u_pixel.isclose((0.01, 0.0, 0.0))
    assert image.dxf.v_pixel.isclose((0.0, 0.01, 0.0))
    assert image.dxf.image_size.isclose((640.0, 360.0))
    assert image.dxf.flags == 7
    assert image.dxf.clipping == 0
    assert image.dxf.brightness == 50
    assert image.dxf.contrast == 50
    assert image.dxf.fade == 0
    assert image.dxf.image_def_reactor_handle == "DEAD"
    assert image.dxf.clipping_boundary_type == 1
    assert image.dxf.count_boundary_points == 2
    x, y, *_ = image.dxf.image_size
    assert image.boundary_path == [(-0.5, -0.5), (x - 0.5, y - 0.5)]


def test_boundary_path(image):
    assert [(-0.5, -0.5), (639.5, 359.5)] == image.boundary_path


def test_reset_boundary_path(image):
    image.reset_boundary_path()
    assert 2 == image.dxf.count_boundary_points
    assert image.get_flag_state(image.USE_CLIPPING_BOUNDARY) is False
    assert image.dxf.clipping == 0
    x, y, *_ = image.dxf.image_size
    assert [(-0.5, -0.5), (x - 0.5, y - 0.5)] == image.boundary_path


def test_set_boundary_path(image):
    image.set_boundary_path(
        [(0, 0), (640, 180), (320, 360)]
    )  # 3 vertices triangle
    assert image.dxf.count_boundary_points == 4
    assert image.dxf.clipping_boundary_type == 2
    # auto close
    assert [(0, 0), (640, 180), (320, 360), (0, 0)] == image.boundary_path
    assert image.dxf.clipping == 1
    assert image.get_flag_state(image.USE_CLIPPING_BOUNDARY) is True


def test_post_load_hook_creates_image_def_reactor(doc):
    image_def = doc.add_image_def("test.jpg", (1, 1))
    msp = doc.modelspace()
    image = msp.add_image(image_def, (0, 0), (1, 1))
    old_reactor_handle = image.dxf.image_def_reactor_handle

    # Hack to unlink Image from ImageDefReactor!
    image.dxf.image_def_reactor_handle = None
    command = image.post_load_hook(doc)

    assert command is not None, "must return a fixer as callable object"
    command()  # execute fix
    handle = image.dxf.image_def_reactor_handle
    reactor = doc.entitydb[handle]

    assert handle is not None, "must have an ImageDefReactor"
    assert handle != old_reactor_handle, "must have a new ImageDefReactor"
    assert (
        handle in image_def.reactors
    ), "ImageDef must be linked to ImageDefReactor"
    assert (
        reactor.dxf.image_handle == image.dxf.handle
    ), "ImageDefReactor must be linked to Image"


def test_exception_while_fixing_image_def_reactor(doc):
    from io import StringIO

    stream = StringIO()
    logging_handler = logging.StreamHandler(stream)
    logger.addHandler(logging_handler)

    image_def = doc.add_image_def("test.jpg", (1, 1))
    msp = doc.modelspace()
    image = msp.add_image(image_def, (0, 0), (1, 1))

    # Hack to unlink Image from ImageDefReactor!
    image.dxf.image_def_reactor_handle = None
    command = image.post_load_hook(doc)

    # Sabotage execution:
    image.doc = None
    try:
        command()
        msg = stream.getvalue()
    finally:
        logger.removeHandler(logging_handler)

    assert image.is_alive is False, "Exception while fixing must destroy IMAGE"
    assert "AttributeError" in msg
    # Example logging message:
    # ------------------------
    # An exception occurred while executing fixing command for IMAGE(#30), destroying entity.
    # Traceback (most recent call last):
    #   File "D:\Source\ezdxf.git\src\ezdxf\entities\image.py", line 173, in _fix_missing_image_def_reactor
    #     self._create_image_def_reactor()
    #   File "D:\Source\ezdxf.git\src\ezdxf\entities\image.py", line 186, in _create_image_def_reactor
    #     image_def_reactor = self.doc.objects.add_image_def_reactor(
    # AttributeError: 'NoneType' object has no attribute 'objects'


def test_post_load_hook_destroys_image_without_valid_image_def(doc):
    image_def = doc.add_image_def("test.jpg", (1, 1))
    msp = doc.modelspace()
    image = msp.add_image(image_def, (0, 0), (1, 1))

    # Hack to unlink Image from ImageDef!
    image.dxf.image_def_handle = None
    image.post_load_hook(doc)

    assert (
        image.is_alive is False
    ), "Image with invalid ImageDef must be destroyed"


@pytest.fixture
def new_doc():
    return ezdxf.new("R2000")


def test_new_image_def(new_doc):
    rootdict = new_doc.rootdict
    assert "ACAD_IMAGE_DICT" not in rootdict
    imagedef = new_doc.add_image_def("mycat.jpg", size_in_pixel=(640, 360))

    # check internals image_def_owner -> ACAD_IMAGE_DICT
    image_dict = rootdict["ACAD_IMAGE_DICT"]
    assert imagedef.dxf.owner == image_dict.dxf.handle

    assert imagedef.dxf.filename == "mycat.jpg"
    assert imagedef.dxf.image_size.isclose((640.0, 360.0))

    # rest are default values
    assert imagedef.dxf.pixel_size.isclose((0.01, 0.01))
    assert imagedef.dxf.loaded == 1
    assert imagedef.dxf.resolution_units == 0


def test_create_and_delete_image(new_doc):
    msp = new_doc.modelspace()
    image_def = new_doc.add_image_def("mycat.jpg", size_in_pixel=(640, 360))
    image = msp.add_image(
        image_def=image_def, insert=(0, 0), size_in_units=(3.2, 1.8)
    )
    assert image.dxf.insert.isclose((0, 0, 0))
    assert image.dxf.u_pixel.isclose((0.005, 0, 0))
    assert image.dxf.v_pixel.isclose((0.0, 0.005, 0))
    assert image.dxf.image_size.isclose((640, 360))
    assert image_def.dxf.handle == image.dxf.image_def_handle
    assert image.dxf.flags == 3
    assert image.dxf.clipping == 0
    assert image.dxf.count_boundary_points == 2
    x, y = image.dxf.image_size.vec2
    assert image.boundary_path == [(-0.5, -0.5), (x - 0.5, y - 0.5)]

    image_def2 = image.image_def
    assert image_def.dxf.handle, image_def2.dxf.handle

    # does image def reactor exists
    reactor_handle = image.dxf.image_def_reactor_handle
    assert reactor_handle in new_doc.objects
    reactor = new_doc.entitydb[reactor_handle]
    assert (
        image.dxf.handle == reactor.dxf.owner
    ), "IMAGE is not owner of IMAGEDEF_REACTOR"
    assert (
        image.dxf.handle == reactor.dxf.image_handle
    ), "IMAGEDEF_REACTOR does not point to IMAGE"

    assert (
        reactor_handle in image_def2.get_reactors()
    ), "Reactor handle not in IMAGE_DEF reactors."

    # delete image
    msp.delete_entity(image)
    new_doc.entitydb.purge()
    assert reactor.is_alive is False
    assert (
        reactor_handle not in new_doc.objects
    ), "IMAGEDEF_REACTOR not deleted for objects section"
    assert (
        reactor_handle not in new_doc.entitydb
    ), "IMAGEDEF_REACTOR not deleted for entity database"
    assert (
        reactor_handle not in image_def2.get_reactors()
    ), "Reactor handle not deleted from IMAGE_DEF reactors."


def create_image(doc):
    msp = doc.modelspace()
    image_def = doc.add_image_def("mycat.jpg", size_in_pixel=(640, 360))
    image = msp.add_image(
        image_def=image_def, insert=(0, 0), size_in_units=(3.2, 1.8)
    )
    image.boundary_path = [(0, 0), (1, 1), (2, 2), (3, 3)]
    return image_def, image


def test_create_and_copy_image(new_doc):
    msp = new_doc.modelspace()
    entitydb = new_doc.entitydb
    image_def, image = create_image(new_doc)
    reactor_handle = image.dxf.image_def_reactor_handle
    copy = image.copy()
    msp.add_entity(copy)
    reactor_handle_of_copy = copy.dxf.image_def_reactor_handle

    # Each image has it's own ImageDefReactor
    assert (
        reactor_handle in entitydb
    ), "ImageDefReactor of source image not stored in EntityDB."
    assert (
        reactor_handle_of_copy in entitydb
    ), "ImageDefReactor of copy not stored in EntityDB."
    assert (
        reactor_handle != reactor_handle_of_copy
    ), "Source image and copy must not have the same ImageDefReactor."

    # Both images are linked to same ImageDef object.
    assert (
        reactor_handle in image_def.reactors
    ), "ImageDefReactor of source image not stored in ImageDef object."
    assert (
        reactor_handle_of_copy in image_def.reactors
    ), "ImageDefReactor of copied image not stored in ImageDef object."
    assert (
        image.boundary_path == copy.boundary_path
    ), "Invalid copy of boundary path."
    assert (
        image.boundary_path is not copy.boundary_path
    ), "Copied image needs an independent copy of the boundary path."


def test_moving_to_another_layout(new_doc):
    msp = new_doc.modelspace()
    image_def, image = create_image(new_doc)
    old_reactor_handle = image.dxf.image_def_reactor_handle
    blk = new_doc.blocks.new("TEST")
    msp.move_to_layout(image, blk)
    # Moving between layouts should not alter the ImageDefReactor
    assert (
        image.dxf.image_def_reactor_handle == old_reactor_handle
    ), "Moved image got a new ImageDefReactor without necessity."


def test_copying_to_another_layout(new_doc):
    image_def, image = create_image(new_doc)
    old_reactor_handle = image.dxf.image_def_reactor_handle
    blk = new_doc.blocks.new("TEST")
    copy = new_doc.entitydb.duplicate_entity(image)

    # duplicate_entity() binds entity to document and triggers the
    # post_bind_hook() to create a new ImageDefReactor:
    copy_reactor_handle = copy.dxf.image_def_reactor_handle
    assert (
        copy_reactor_handle != "0"
    ), "Image copy did not get a new ImageDefReactor"
    assert (
        copy_reactor_handle != old_reactor_handle
    ), "Image copy has the same ImageDefReactor as the source image."

    blk.add_entity(copy)

    # unlinking an image, does not remove the ImageDefReactor
    blk.unlink_entity(copy)
    assert (
        copy.dxf.image_def_reactor_handle == copy_reactor_handle
    ), "Removed ImageDefReactor from unlinked image."

    # Keep reactor if adding an image with existing ImageDefReactor
    blk.add_entity(copy)
    assert (
        copy.dxf.image_def_reactor_handle == copy_reactor_handle
    ), "Relinked image got new ImageDefReactor without necessity."


IMAGE_DEF = """  0
IMAGEDEF
  5
DEAD
330
DEAD
100
AcDbRasterImageDef
 90
        0
  1
mycat.jpg
 10
640.0
 20
360.0
 11
0.01
 21
0.01
280
     1
281
     0
"""

IMAGE = """  0
IMAGE
  5
1DF
330
1F
100
AcDbEntity
  8
0
100
AcDbRasterImage
 90
        0
 10
0.0
 20
0.0
 30
0.0
 11
0.01
 21
0.0
 31
0.0
 12
0.0
 22
0.01
 32
0.0
 13
640.0
 23
360.0
340
DEAD
 70
     7
280
     0
281
    50
282
    50
283
     0
360
DEAD
 71
     1
 91
        2
 14
-.5
 24
-.5
 14
639.5
 24
359.5
"""



# Copyright (c) 2016-2017, Manfred Moitzi
# License: MIT License
import pytest
import ezdxf

_OBJECT_TABLE_NAMES = [
    "ACAD_COLOR",
    "ACAD_GROUP",
    "ACAD_LAYOUT",
    "ACAD_MATERIAL",
    "ACAD_MLEADERSTYLE",
    "ACAD_MLINESTYLE",
    "ACAD_PLOTSETTINGS",
    "ACAD_PLOTSTYLENAME",
    "ACAD_SCALELIST",
    "ACAD_TABLESTYLE",
    "ACAD_VISUALSTYLE",
]


@pytest.fixture(scope="module")
def doc():
    return ezdxf.new("R2000")


def test_setup_rootdict(doc):
    rootdict = doc.rootdict
    assert "DICTIONARY" == rootdict.dxftype()


def test_add_new_sub_dict(doc):
    rootdict = doc.rootdict
    new_dict = rootdict.add_new_dict("A_SUB_DICT")
    assert "DICTIONARY" == new_dict.dxftype()
    assert 0 == len(new_dict)
    assert "A_SUB_DICT" in rootdict
    assert rootdict.dxf.handle == new_dict.dxf.owner


def test_required_tables_exists(doc):
    rootdict = doc.rootdict
    for table_name in _OBJECT_TABLE_NAMES:
        assert table_name in rootdict, "table %s not found." % table_name


def test_new_plot_style_name_table(doc):
    rootdict = doc.rootdict
    plot_style_name_table = rootdict["ACAD_PLOTSTYLENAME"]
    assert "ACDBDICTIONARYWDFLT" == plot_style_name_table.dxftype()
    place_holder = plot_style_name_table["Normal"]
    assert "ACDBPLACEHOLDER" == place_holder.dxftype()
    assert place_holder.dxf.owner == plot_style_name_table.dxf.handle



# Copyright (c) 2010-2022, Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import TYPE_CHECKING, Any
import math
import enum


from ezdxf.enums import (
    MTextEntityAlignment,
    MAP_MTEXT_ALIGN_TO_FLAGS,
)
from ezdxf.lldxf import const
from ezdxf.math import UVec, Vec3
from .mixins import SubscriptAttributes

if TYPE_CHECKING:
    from ezdxf.eztypes import GenericLayoutType

TOP_ALIGN = {
    MTextEntityAlignment.TOP_LEFT,
    MTextEntityAlignment.TOP_RIGHT,
    MTextEntityAlignment.TOP_CENTER,
}
MIDDLE_ALIGN = {
    MTextEntityAlignment.MIDDLE_LEFT,
    MTextEntityAlignment.MIDDLE_CENTER,
    MTextEntityAlignment.MIDDLE_RIGHT,
}


class Mirror(enum.IntEnum):
    NONE = 0
    MIRROR_X = 2
    MIRROR_Y = 4


class MTextSurrogate(SubscriptAttributes):
    """MTEXT surrogate for DXF R12 build up by TEXT Entities. This add-on was
    added to simplify the transition from :mod:`dxfwrite` to :mod:`ezdxf`.

    The rich-text formatting capabilities for the regular MTEXT entity are not
    supported, if these features are required use the regular MTEXT entity and
    the :class:`~ezdxf.addons.MTextExplode` add-on to explode the MTEXT entity
    into DXF primitives.

    .. important::

        The align-point is always the insert-point, there is no need for
        a second align-point because the horizontal alignments FIT, ALIGN,
        BASELINE_MIDDLE are not supported.

    Args:
        text: content as string
        insert: insert location in drawing units
        line_spacing: line spacing in percent of height, 1.5 = 150% = 1+1/2 lines
        align: text alignment as :class:`~ezdxf.enums.MTextEntityAlignment` enum
        char_height: text height in drawing units
        style: :class:`~ezdxf.entities.Textstyle` name as string
        oblique: oblique angle in degrees, where 0 is vertical
        rotation: text rotation angle in degrees
        width_factor: text width factor as float
        mirror: :attr:`MTextSurrogate.MIRROR_X` to mirror the text horizontal
            or :attr:`MTextSurrogate.MIRROR_Y` to mirror the text vertical
        layer: layer name as string
        color: :ref:`ACI`

    """

    MIRROR_NONE = Mirror.NONE
    MIRROR_X = Mirror.MIRROR_X
    MIRROR_Y = Mirror.MIRROR_Y

    def __init__(
        self,
        text: str,
        insert: UVec,
        line_spacing: float = 1.5,
        align=MTextEntityAlignment.TOP_LEFT,
        char_height: float = 1.0,
        style="STANDARD",
        oblique: float = 0.0,
        rotation: float = 0.0,
        width_factor: float = 1.0,
        mirror=Mirror.NONE,
        layer="0",
        color: int = const.BYLAYER,
    ):
        self.content: list[str] = text.split("\n")
        self.insert = Vec3(insert)
        self.line_spacing = float(line_spacing)
        assert isinstance(align, MTextEntityAlignment)
        self.align = align

        self.char_height = float(char_height)
        self.style = str(style)
        self.oblique = float(oblique)  # in degree
        self.rotation = float(rotation)  # in degree
        self.width_factor = float(width_factor)
        self.mirror = int(mirror)  # renamed to text_generation_flag in ezdxf
        self.layer = str(layer)
        self.color = int(color)

    @property
    def line_height(self) -> float:
        """Absolute line spacing in drawing units."""
        return self.char_height * self.line_spacing

    def render(self, layout: GenericLayoutType) -> None:
        """Render the multi-line content as separated TEXT entities into the
        given `layout` instance.
        """
        text_lines = self.content
        if len(text_lines) > 1:
            if self.mirror & const.MIRROR_Y:
                text_lines.reverse()
            for line_number, text in enumerate(text_lines):
                align_point = self._get_align_point(line_number)
                layout.add_text(
                    text,
                    dxfattribs=self._dxfattribs(align_point),
                )
        elif len(text_lines) == 1:
            layout.add_text(
                text_lines[0],
                dxfattribs=self._dxfattribs(self.insert),
            )

    def _get_align_point(self, line_number: int) -> Vec3:
        """Calculate the align-point depending on the line number."""
        x = self.insert.x
        y = self.insert.y
        try:
            z = self.insert.z
        except IndexError:
            z = 0.0
        # rotation not respected

        if self.align in TOP_ALIGN:
            y -= line_number * self.line_height
        elif self.align in MIDDLE_ALIGN:
            y0 = line_number * self.line_height
            full_height = (len(self.content) - 1) * self.line_height
            y += (full_height / 2) - y0
        else:  # BOTTOM ALIGN
            y += (len(self.content) - 1 - line_number) * self.line_height
        return self._rotate(Vec3(x, y, z))

    def _rotate(self, alignpoint: Vec3) -> Vec3:
        """Rotate `alignpoint` around insert-point about rotation degrees."""
        dx = alignpoint.x - self.insert.x
        dy = alignpoint.y - self.insert.y
        beta = math.radians(self.rotation)
        x = self.insert.x + dx * math.cos(beta) - dy * math.sin(beta)
        y = self.insert.y + dy * math.cos(beta) + dx * math.sin(beta)
        return Vec3(round(x, 6), round(y, 6), alignpoint.z)

    def _dxfattribs(self, align_point: Vec3) -> dict[str, Any]:
        """Build keyword arguments for TEXT entity creation."""
        halign, valign = MAP_MTEXT_ALIGN_TO_FLAGS[self.align]
        return {
            "insert": align_point,
            "align_point": align_point,
            "layer": self.layer,
            "color": self.color,
            "style": self.style,
            "height": self.char_height,
            "width": self.width_factor,
            "text_generation_flag": self.mirror,
            "rotation": self.rotation,
            "oblique": self.oblique,
            "halign": halign,
            "valign": valign,
        }



# Copyright (c) 2021-2022, Manfred Moitzi
# License: MIT License
from __future__ import annotations
from ezdxf.addons.xqt import QTableView, QTreeView, QModelIndex
from ezdxf.lldxf.tags import Tags


class StructureTree(QTreeView):
    def set_structure(self, model):
        self.setModel(model)
        self.expand(model.index(0, 0, QModelIndex()))
        self.setHeaderHidden(True)

    def expand_to_entity(self, entity: Tags):
        model = self.model()
        index = model.index_of_entity(entity)  # type: ignore
        self.setCurrentIndex(index)


class DXFTagsTable(QTableView):
    def __init__(self):
        super().__init__()
        col_header = self.horizontalHeader()
        col_header.setStretchLastSection(True)
        row_header = self.verticalHeader()
        row_header.setDefaultSectionSize(24)  # default row height in pixels
        self.setSelectionBehavior(QTableView.SelectRows)

    def first_selected_row(self) -> int:
        first_row: int = 0
        selection = self.selectedIndexes()
        if selection:
            first_row = selection[0].row()
        return first_row

    def selected_rows(self) -> list[int]:
        rows: set[int] = set()
        selection = self.selectedIndexes()
        for item in selection:
            rows.add(item.row())
        return sorted(rows)



# Copyright (c) 2024, Manfred Moitzi
# License: MIT License
import pytest
import ezdxf

from ezdxf.document import Drawing
from ezdxf.math import Matrix44
from ezdxf.render.forms import cube
from ezdxf.acis.api import body_from_mesh, export_sab

FILENAME = "3dsolid.dxf"


@pytest.fixture
def doc():
    acis_body = body_from_mesh(cube())
    _doc = ezdxf.new()
    msp = _doc.modelspace()
    dxf_3dsolid = msp.add_3dsolid()
    dxf_3dsolid.sab = export_sab([acis_body])
    dxf_3dsolid.transform(Matrix44.translate(10, 11, 12))
    return _doc


def test_apply_transformations_to_acis_entities(doc: Drawing):
    # this is done automatically when exporting the DXF document
    try:
        doc.commit_pending_changes()
    except Exception as e:
        pytest.fail(f"commit_pending_changes() raises unexpected exception: {str(e)}")


def test_export_and_reload_transformed_acis_entities(tmp_path, doc: Drawing):
    file_path = tmp_path / FILENAME
    doc.saveas(file_path)
    del doc

    doc2 = ezdxf.readfile(file_path)
    # Do not test the current implementation, this may change in the future:
    solids = doc2.query("3DSOLID")
    assert len(solids) == 1

    dxf_3dsolid = solids[0]
    assert len(dxf_3dsolid.sab) > 16


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2019-2021 Manfred Moitzi
# License: MIT License
import pytest

from ezdxf.entities.hatch import Hatch
from ezdxf.lldxf.tagwriter import TagCollector, basic_tags_from_text
from ezdxf.lldxf import const
from ezdxf.math import Vec3


HATCH = """0
HATCH
5
0
330
0
100
AcDbEntity
8
0
62
1
100
AcDbHatch
10
0.0
20
0.0
30
0.0
210
0.0
220
0.0
230
1.0
2
SOLID
70
1
71
0
91
0
75
1
76
1
98
1
10
0.0
20
0.0
"""


@pytest.fixture
def entity() -> Hatch:
    return Hatch.from_text(HATCH)


def test_if_registered():
    from ezdxf.entities.factory import ENTITY_CLASSES

    assert "HATCH" in ENTITY_CLASSES


def test_default_init():
    entity = Hatch()
    assert entity.dxftype() == "HATCH"
    assert entity.dxf.handle is None
    assert entity.dxf.owner is None


def test_default_new():
    entity = Hatch.new(
        handle="ABBA",
        owner="0",
        dxfattribs={
            "color": 7,
        },
    )
    assert entity.dxf.layer == "0"
    assert entity.dxf.color == 7


def test_load_from_text(entity):
    assert entity.dxf.layer == "0"
    assert entity.dxf.color == 1, "default color is 1"


def test_seeds(entity: Hatch):
    # seeds is a list of (x, y) tuples
    assert entity.seeds == [(0, 0)]


def test_write_dxf():
    entity = Hatch.from_text(HATCH)
    result = TagCollector.dxftags(entity)
    expected = basic_tags_from_text(HATCH)
    assert result == expected


def test_write_correct_polyline_path_tag_order(entity):
    entity = Hatch.from_text(HATCH)
    entity.paths.add_polyline_path(
        [(0, 0), (1, 0), (1, 1)],
        is_closed=True,
    )
    result = TagCollector.dxftags(entity)
    tags = list(result.pop_tags([92, 72, 73]))
    # 92 = path type 3: external polyline path
    # 72 = has_bulge
    # 73 = is_closed
    # The group codes 72 and 73 are swapped in comparison to MPOLYGON
    assert tags == [(92, 3), (72, 0), (73, 1)]


def test_hatch_boundary_state():
    state = const.BoundaryPathState.from_flags(
        const.BOUNDARY_PATH_EXTERNAL
        + const.BOUNDARY_PATH_DERIVED
        + const.BOUNDARY_PATH_TEXTBOX
        + const.BOUNDARY_PATH_OUTERMOST
    )
    assert state.external is True
    assert state.derived is True
    assert state.textbox is True
    assert state.outermost is True
    assert state.default is False


def test_hatch_boundary_default_state():
    state = const.BoundaryPathState()
    assert state.external is False
    assert state.derived is False
    assert state.textbox is False
    assert state.outermost is False
    assert state.default is True




#  Copyright (c) 2021-2022, Manfred Moitzi
#  License: MIT License
#
# Ezdxf user: DO NOT IMPORT FROM ezdxf.path.* MODULES!
# import all Path() related classes and functions from ezdxf.path
#
# from ezdxf.path import Path, make_path,
#
from .commands import *
from .path import *
from .converter import *
from .tools import *
from .nesting import *
from .shapes import *



# Purpose: acdsdata section manager
# Copyright (c) 2014-2022, Manfred Moitzi
# License: MIT License
"""
ACDSDATA entities have NO handles, therefore they can not be stored in the
drawing entity database.

section structure (work in progress):
0 <str> SECTION
2 <str> ACDSDATA
70 <int> 2 # flag?
71 <int> 6 # count of following ACDSSCHEMA entities ??? no, just another flag

0 <str> ACDSSCHEMA           # dxftype: schema definition
90 <int> 0                   # schema number 0, 1, 2, 3 ...
1 <str> AcDb3DSolid_ASM_Data # schema name

2 <str> AcDbDs::ID           # subsection name
280 <int> 10                 # subsection type 10 = ???
91 <int> 8                   # data ???

2 <str> ASM_Data             # subsection name
280 <int> 15                 # subsection type
91 <int> 0                   # data ???
101 <str> ACDSRECORD         # data
95 <int> 0
90 <int> 2
...

0 <str> ACDSSCHEMA
90 <int> 1
1 <str> AcDb_Thumbnail_Schema
...

0 <str> ACDSSCHEMA
90 <int> 2
1 <str> AcDbDs::TreatedAsObjectDataSchema
...

0 <str> ACDSSCHEMA
90 <int> 3
1 <str> AcDbDs::LegacySchema
2 <str> AcDbDs::Legacy
280 <int> 1
91 <int> 0

0 <str> ACDSSCHEMA
90 <int> 4
1 <str> AcDbDs::IndexedPropertySchema
2 <str> AcDs:Indexable
280 <int> 1
91 <int> 0

0 <str> ACDSSCHEMA
90 <int> 5
1 <str> AcDbDs::HandleAttributeSchema
2 <str> AcDbDs::HandleAttribute
280 <int> 7
91 <int> 1
284 <int> 1

0 <str> ACDSRECORD               # dxftype: data record
90 <int> 0                       # ??? flag
2 <str> AcDbDs::ID               # subsection name
280 <int> 10                     # subsection type 10 = handle to owner entity, 3DSOLID/REGION
320 <str> 339                    # handle
2 <str> ASM_Data                 # subsection name
280 <int> 15                     # subsection type 15 = binary data
94 <int> 1088                    # size of data
310 <binary encoded data>        # data
310 <binary encoded data>        # data
...

0 <str> ENDSEC
"""
from __future__ import annotations
from typing import TYPE_CHECKING, Iterator, Iterable, Any, Optional
import abc
from itertools import islice

from ezdxf.lldxf.tags import group_tags, Tags
from ezdxf.lldxf.types import dxftag
from ezdxf.lldxf.const import DXFKeyError, DXFStructureError

if TYPE_CHECKING:
    from ezdxf.document import Drawing
    from ezdxf.lldxf.tagwriter import AbstractTagWriter


class AcDsEntity(abc.ABC):
    @abc.abstractmethod
    def export_dxf(self, tagwriter: AbstractTagWriter):
        ...

    @abc.abstractmethod
    def dxftype(self) -> str:
        ...


class AcDsDataSection:
    name = "ACDSDATA"

    def __init__(self, doc: Drawing, entities: Optional[Iterable[Tags]] = None):
        self.doc = doc
        self.entities: list[AcDsEntity] = []
        self.section_info = Tags()
        if entities is not None:
            self.load_tags(iter(entities))

    @property
    def is_valid(self) -> bool:
        return len(self.section_info) > 0

    @property
    def has_records(self) -> bool:
        return any(
            isinstance(entity, AcDsRecord) for entity in self.entities
        )

    def load_tags(self, entities: Iterator[Tags]) -> None:
        section_head = next(entities)
        if section_head[0] != (0, "SECTION") or section_head[1] != (
            2,
            "ACDSDATA",
        ):
            raise DXFStructureError(
                "Critical structure error in ACDSDATA section."
            )

        self.section_info = section_head
        for entity in entities:
            self.append(AcDsData(entity))  # tags have no subclasses

    def append(self, entity: AcDsData) -> None:
        cls = ACDSDATA_TYPES.get(entity.dxftype(), AcDsData)
        data = cls(entity.tags)
        self.entities.append(data)

    def export_dxf(self, tagwriter: AbstractTagWriter) -> None:
        if not self.is_valid or not self.has_records:
            # Empty ACDSDATA section is not required!
            return
        tagwriter.write_tags(self.section_info)
        for entity in self.entities:
            entity.export_dxf(tagwriter)
        tagwriter.write_tag2(0, "ENDSEC")

    @property
    def acdsrecords(self) -> Iterator[AcDsRecord]:
        return (
            entity for entity in self.entities if isinstance(entity, AcDsRecord)
        )

    def get_acis_data(self, handle: str) -> bytes:
        asm_record = self.find_acis_record(handle)
        if asm_record is not None:
            return b"".join(get_acis_data(asm_record))
        return b""

    def set_acis_data(self, handle: str, sab_data: bytes) -> None:
        asm_record = self.find_acis_record(handle)
        if asm_record is not None:
            set_acis_data(asm_record, sab_data)
        else:
            self.new_acis_data(handle, sab_data)

    def new_acis_data(self, handle: str, sab_data: bytes) -> None:
        self.entities.append(new_acis_record(handle, sab_data))

    def del_acis_data(self, handle) -> None:
        asm_record = self.find_acis_record(handle)
        if asm_record is not None:
            self.entities.remove(asm_record)

    def find_acis_record(self, handle: str) -> Optional[AcDsRecord]:
        for record in self.acdsrecords:
            if is_acis_data(record) and acis_entity_handle(record) == handle:
                return record
        return None


class AcDsData(AcDsEntity):
    def __init__(self, tags: Tags):
        self.tags = tags

    def export_dxf(self, tagwriter: AbstractTagWriter):
        tagwriter.write_tags(self.tags)

    def dxftype(self) -> str:
        return self.tags[0].value


class Section(Tags):
    @property
    def name(self) -> str:
        return self[0].value

    @property
    def type(self) -> str:
        return self[1].value

    @property
    def data(self) -> Tags:
        return Tags(self[2:])


class AcDsRecord(AcDsEntity):
    def __init__(self, tags: Tags):
        self._dxftype = tags[0]
        self.flags = tags[1]
        self.sections = [
            Section(group)
            for group in group_tags(islice(tags, 2, None), splitcode=2)
        ]

    def dxftype(self) -> str:
        return "ACDSRECORD"

    def has_section(self, name: str) -> bool:
        return self.get_section(name, default=None) is not None

    def get_section(self, name: str, default: Any = DXFKeyError) -> Section:
        for section in self.sections:
            if section.name == name:
                return section
        if default is DXFKeyError:
            raise DXFKeyError(name)
        else:
            return default

    def index(self, name: str) -> int:
        for i, section in enumerate(self.sections):
            if section.name == name:
                return i
        return -1

    def replace(self, section: Section) -> None:
        index = self.index(section.name)
        if index == -1:
            self.sections.append(section)
        else:
            self.sections[index] = section

    def append(self, section: Section):
        self.sections.append(section)

    def __len__(self):
        return len(self.sections)

    def __getitem__(self, item) -> Section:
        return self.sections[item]

    def _write_header(self, tagwriter: AbstractTagWriter) -> None:
        tagwriter.write_tags(Tags([self._dxftype, self.flags]))

    def export_dxf(self, tagwriter: AbstractTagWriter) -> None:
        self._write_header(tagwriter)
        for section in self.sections:
            tagwriter.write_tags(section)


def get_acis_data(record: AcDsRecord) -> list[bytes]:
    try:
        asm_data = record.get_section("ASM_Data")
    except DXFKeyError:  # no data stored
        return []
    else:
        return [tag.value for tag in asm_data if tag.code == 310]


def is_acis_data(record: AcDsRecord) -> bool:
    return record.has_section("ASM_Data")


def acis_entity_handle(record: AcDsRecord) -> str:
    try:
        section = record.get_section("AcDbDs::ID")
    except DXFKeyError:  # not present
        return ""
    return section.get_first_value(320, "")


def set_acis_data(record: AcDsRecord, data: bytes) -> None:
    chunk_size = 127
    size = len(data)
    tags = Tags(
        [
            dxftag(2, "ASM_Data"),
            dxftag(280, 15),
            dxftag(94, size),
        ]
    )
    index = 0
    while index < size:
        tags.append(dxftag(310, data[index : index + chunk_size]))
        index += chunk_size
    record.replace(Section(tags))


def new_acis_record(handle: str, sab_data: bytes) -> AcDsRecord:
    tags = Tags(
        [
            dxftag(0, "ACDSRECORD"),
            dxftag(90, 1),
            dxftag(2, "AcDbDs::ID"),
            dxftag(280, 10),
            dxftag(320, handle),
        ]
    )
    record = AcDsRecord(tags)
    set_acis_data(record, sab_data)
    return record


ACDSDATA_TYPES = {
    "ACDSRECORD": AcDsRecord,
}


DEFAULT_SETUP = """0
SECTION
2
ACDSDATA
70
2
71
2
0
ACDSSCHEMA
90
0
1
AcDb_Thumbnail_Schema
2
AcDbDs::ID
280
10
91
8
2
Thumbnail_Data
280
15
91
0
101
ACDSRECORD
95
0
90
2
2
AcDbDs::TreatedAsObjectData
280
1
291
1
101
ACDSRECORD
95
0
90
3
2
AcDbDs::Legacy
280
1
291
1
101
ACDSRECORD
1
AcDbDs::ID
90
4
2
AcDs:Indexable
280
1
291
1
101
ACDSRECORD
1
AcDbDs::ID
90
5
2
AcDbDs::HandleAttribute
280
7
282
1
0
ACDSSCHEMA
90
1
1
AcDb3DSolid_ASM_Data
2
AcDbDs::ID
280
10
91
8
2
ASM_Data
280
15
91
0
101
ACDSRECORD
95
1
90
2
2
AcDbDs::TreatedAsObjectData
280
1
291
1
101
ACDSRECORD
95
1
90
3
2
AcDbDs::Legacy
280
1
291
1
101
ACDSRECORD
1
AcDbDs::ID
90
4
2
AcDs:Indexable
280
1
291
1
101
ACDSRECORD
1
AcDbDs::ID
90
5
2
AcDbDs::HandleAttribute
280
7
282
1
0
ACDSSCHEMA
90
2
1
AcDbDs::TreatedAsObjectDataSchema
2
AcDbDs::TreatedAsObjectData
280
1
91
0
0
ACDSSCHEMA
90
3
1
AcDbDs::LegacySchema
2
AcDbDs::Legacy
280
1
91
0
0
ACDSSCHEMA
90
4
1
AcDbDs::IndexedPropertySchema
2
AcDs:Indexable
280
1
91
0
0
ACDSSCHEMA
90
5
1
AcDbDs::HandleAttributeSchema
2
AcDbDs::HandleAttribute
280
7
91
1
284
1
"""
# (0, ENDSEC) must be omitted!


def new_acds_data_section(doc: Drawing) -> AcDsDataSection:
    if doc.dxfversion >= "AC1027":
        return AcDsDataSection(doc, group_tags(Tags.from_text(DEFAULT_SETUP)))
    else:
        return AcDsDataSection(doc)



# Copyright (c) 2019-2022 Manfred Moitzi
# License: MIT License
import pytest
import math

from ezdxf.entities.line import Line
from ezdxf.lldxf.const import DXF12, DXF2000, DXFValueError
from ezdxf.lldxf.tagwriter import TagCollector, basic_tags_from_text
from ezdxf.math import Matrix44, Z_AXIS, Y_AXIS

TEST_CLASS = Line
TEST_TYPE = "LINE"

ENTITY_R12 = """0
LINE
5
0
8
0
10
0.0
20
0.0
30
0.0
11
1.0
21
1.0
31
1.0
"""

ENTITY_R2000 = """0
LINE
5
0
330
0
100
AcDbEntity
8
0
100
AcDbLine
10
0.0
20
0.0
30
0.0
11
1.0
21
1.0
31
1.0
"""


@pytest.fixture(params=[ENTITY_R12, ENTITY_R2000])
def line(request):
    return Line.from_text(request.param)


def test_registered():
    from ezdxf.entities.factory import ENTITY_CLASSES

    assert TEST_TYPE in ENTITY_CLASSES


def test_default_init():
    entity = Line()
    assert entity.dxftype() == TEST_TYPE


def test_default_new():
    entity = Line.new(
        handle="ABBA",
        owner="0",
        dxfattribs={
            "color": "7",
            "start": (1, 2, 3),
            "end": (4, 5, 6),
        },
    )
    assert entity.dxf.layer == "0"
    assert entity.dxf.color == 7
    assert entity.dxf.start == (1, 2, 3)
    assert entity.dxf.start.x == 1, "is not Vec3 compatible"
    assert entity.dxf.start.y == 2, "is not Vec3 compatible"
    assert entity.dxf.start.z == 3, "is not Vec3 compatible"
    assert entity.dxf.end == (4, 5, 6)
    assert entity.dxf.extrusion == (0.0, 0.0, 1.0)
    assert entity.dxf.hasattr("extrusion") is False, "just the default value"

    # can set DXF R2007 value
    entity.dxf.shadow_mode = 1
    assert entity.dxf.shadow_mode == 1


def test_load_from_text(line):
    assert line.dxf.layer == "0"
    assert line.dxf.color == 256, "default color is 256 (by layer)"
    assert line.dxf.start == (0, 0, 0)
    assert line.dxf.end == (1, 1, 1)


@pytest.mark.parametrize(
    "txt,ver", [(ENTITY_R2000, DXF2000), (ENTITY_R12, DXF12)]
)
def test_write_dxf(txt, ver):
    expected = basic_tags_from_text(txt)
    line = TEST_CLASS.from_text(txt)
    collector = TagCollector(dxfversion=ver, optional=True)
    line.export_dxf(collector)
    assert collector.tags == expected

    collector2 = TagCollector(dxfversion=ver, optional=False)
    line.export_dxf(collector2)
    assert collector.has_all_tags(collector2)


def test_get_pass_through_ocs():
    line = Line.new(
        dxfattribs={
            "start": (0, 0, 0),
            "end": (1, 0, 0),
            "extrusion": (0, 0, -1),
        }
    )
    ocs = line.ocs()
    assert ocs.to_wcs((0, 0, 0)) == (0, 0, 0)
    assert ocs.to_wcs((1, 0, 0)) == (1, 0, 0)


def test_transform():
    line = Line.new(
        dxfattribs={
            "start": (0, 0, 0),
            "end": (1, 0, 0),
            "extrusion": (0, 1, 0),
        }
    )
    m = Matrix44.translate(1, 2, 3)
    line.transform(m)

    # simple 3D entity - no OCS transformation,
    assert line.dxf.start == (1, 2, 3)
    assert line.dxf.end == (2, 2, 3)
    # extrusion direction without translation - not an OCS extrusion vector!
    assert line.dxf.extrusion == Y_AXIS

    # Create new entity by transformation:
    new_line = line.copy()
    new_line.transform(m)

    assert new_line.dxf.start == (2, 4, 6)
    assert new_line.dxf.end == (3, 4, 6)
    assert Y_AXIS.isclose(new_line.dxf.extrusion)


def test_transform_thickness_of_0():
    line = Line.new(
        dxfattribs={
            "start": (0, 0, 0),
            "end": (1, 0, 0),
            "thickness": 0,
        }
    )
    m = Matrix44.translate(1, 2, 3)
    line.transform(m)
    assert line.dxf.thickness == 0


def test_transform_invalid_extrusion_vector():
    line = Line.new(
        dxfattribs={
            "start": (0, 0, 0),
            "end": (1, 0, 0),
            "extrusion": (0, 0, 0),
        }
    )
    m = Matrix44.translate(1, 2, 3)
    line.transform(m)
    assert Z_AXIS.isclose(line.dxf.extrusion)


def test_translation():
    line = Line.new(
        dxfattribs={
            "start": (0, 0, 0),
            "end": (1, 0, 0),
            "extrusion": (0, 1, 0),
        }
    )
    line.translate(1, 2, 3)
    assert line.dxf.start == (1, 2, 3)
    assert line.dxf.end == (2, 2, 3)


def test_rotation():
    line = Line.new(
        dxfattribs={
            "start": (0, 0, 0),
            "end": (1, 0, 0),
            "extrusion": (0, 1, 0),
        }
    )
    angle = math.pi / 4
    m = Matrix44.z_rotate(angle)
    line.transform(m)
    assert line.dxf.start == (0, 0, 0)
    assert line.dxf.end.isclose(
        (math.cos(angle), math.sin(angle), 0), abs_tol=1e-9
    )
    assert line.dxf.extrusion.isclose(
        (-math.cos(angle), math.sin(angle), 0), abs_tol=1e-9
    )
    assert line.dxf.thickness == 0


def test_scaling():
    line = Line.new(
        dxfattribs={
            "start": (0, 0, 0),
            "end": (1, 0, 0),
            "extrusion": (0, 1, 0),
            "thickness": 2,
        }
    )
    m = Matrix44.scale(2, 2, 0)
    line.transform(m)
    assert line.dxf.start == (0, 0, 0)
    assert line.dxf.end == (2, 0, 0)
    assert line.dxf.extrusion == (0, 1, 0)
    assert line.dxf.thickness == 4


def test_copy_entity_transparency():
    line = Line()
    line2 = line.copy()
    assert line2.dxf.hasattr("transparency") is False

    line.transparency = 0.5
    line2 = line.copy()
    assert line2.dxf.transparency == 0x0200007F


def test_setting_invalid_transparency_value_raises_exception():
    line = Line()
    with pytest.raises(DXFValueError):
        line.dxf.transparency = 0


def test_load_entity_with_invalid_transparency():
    line = Line.from_text(ENTITY_INVALID_TRANSPARENCY)
    # No auto fix in normal loading mode - Auditor fixes this issue at DXF
    # attribute level!
    assert line.dxf.transparency == 268435456
    assert (
        line.transparency == 0.0
    ), "should replace invalid transparency by opaque"


ERR_LINE = """0
LINE
5
0
330
0
100
AcDbEntity
100
AcDbLine
8
0
62
1
6
Linetype
10
0.0
20
0.0
30
0.0
11
1.0
21
1.0
31
1.0
"""

ENTITY_INVALID_TRANSPARENCY = """0
LINE
5
0
330
0
100
AcDbEntity
8
0
440
268435456
100
AcDbLine
10
0.0
20
0.0
30
0.0
11
1.0
21
1.0
31
1.0
"""


def test_recover_acdb_entity_tags():
    line = Line.from_text(ERR_LINE)
    assert line.dxf.layer == "0"
    assert line.dxf.color == 1
    assert line.dxf.linetype == "Linetype"


MALFORMED_LINE = """0
LINE
5
0
62
7
330
0
6
LT_EZDXF
8
LY_EZDXF
100
AcDbEntity
10
1.0
20
1.0
30
1.0
100
AcDbLine
11
2.0
21
2.0
31
2.0
100
AcDbInvalidSubclass
"""


def test_malformed_line():
    line = Line.from_text(MALFORMED_LINE)
    assert line.dxf.layer == "LY_EZDXF"
    assert line.dxf.linetype == "LT_EZDXF"
    assert line.dxf.color == 7
    assert line.dxf.start.isclose((1, 1, 1))
    assert line.dxf.end.isclose((2, 2, 2))




# Copyright (c) 2019-2024 Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import TYPE_CHECKING, Union, Iterable, Optional
from typing_extensions import Self
from ezdxf.lldxf import validator
from ezdxf.lldxf.attributes import (
    DXFAttr,
    DXFAttributes,
    DefSubclass,
    XType,
    RETURN_DEFAULT,
    group_code_mapping,
)
from ezdxf.lldxf.const import SUBCLASS_MARKER, DXF2000, DXFTypeError
from ezdxf.lldxf import const
from ezdxf.lldxf.tags import Tags
from ezdxf.math import NULLVEC, Z_AXIS, UVec, Matrix44, Vec3
from .dxfentity import base_class, SubclassProcessor, DXFEntity
from .dxfgfx import DXFGraphic, acdb_entity
from .dxfobj import DXFObject
from .factory import register_entity
from .copy import default_copy
from ezdxf.math.transformtools import (
    InsertTransformationError,
    InsertCoordinateSystem,
)

if TYPE_CHECKING:
    from ezdxf.document import Drawing
    from ezdxf.entities import DXFNamespace
    from ezdxf.lldxf.tagwriter import AbstractTagWriter
    from ezdxf import xref


__all__ = [
    "PdfUnderlay",
    "DwfUnderlay",
    "DgnUnderlay",
    "PdfDefinition",
    "DgnDefinition",
    "DwfDefinition",
    "Underlay",
    "UnderlayDefinition",
]

acdb_underlay = DefSubclass(
    "AcDbUnderlayReference",
    {
        # Hard reference to underlay definition object
        "underlay_def_handle": DXFAttr(340),
        "insert": DXFAttr(10, xtype=XType.point3d, default=NULLVEC),
        # Scale x factor:
        "scale_x": DXFAttr(
            41,
            default=1,
            validator=validator.is_not_zero,
            fixer=RETURN_DEFAULT,
        ),
        # Scale y factor:
        "scale_y": DXFAttr(
            42,
            default=1,
            validator=validator.is_not_zero,
            fixer=RETURN_DEFAULT,
        ),
        # Scale z factor:
        "scale_z": DXFAttr(
            43,
            default=1,
            validator=validator.is_not_zero,
            fixer=RETURN_DEFAULT,
        ),
        # Rotation angle in degrees:
        "rotation": DXFAttr(50, default=0),
        "extrusion": DXFAttr(
            210,
            xtype=XType.point3d,
            default=Z_AXIS,
            optional=True,
            validator=validator.is_not_null_vector,
            fixer=RETURN_DEFAULT,
        ),
        # Underlay display properties:
        # 1 = Clipping is on
        # 2 = Underlay is on
        # 4 = Monochrome
        # 8 = Adjust for background
        "flags": DXFAttr(280, default=10),
        # Contrast value (20-100; default = 100)
        "contrast": DXFAttr(
            281,
            default=100,
            validator=validator.is_in_integer_range(20, 101),
            fixer=validator.fit_into_integer_range(20, 101),
        ),
        # Fade value (0-80; default = 0)
        "fade": DXFAttr(
            282,
            default=0,
            validator=validator.is_in_integer_range(0, 81),
            fixer=validator.fit_into_integer_range(0, 81),
        ),
    },
)
acdb_underlay_group_codes = group_code_mapping(acdb_underlay)


class Underlay(DXFGraphic):
    """Virtual UNDERLAY entity."""

    # DXFTYPE = 'UNDERLAY'
    DXFATTRIBS = DXFAttributes(base_class, acdb_entity, acdb_underlay)
    MIN_DXF_VERSION_FOR_EXPORT = DXF2000

    def __init__(self) -> None:
        super().__init__()
        self._boundary_path: list[UVec] = []
        self._underlay_def: Optional[UnderlayDefinition] = None

    def copy_data(self, entity: Self, copy_strategy=default_copy) -> None:
        assert isinstance(entity, Underlay)
        entity._boundary_path = list(self._boundary_path)
        entity._underlay_def = self._underlay_def

    def load_dxf_attribs(
        self, processor: Optional[SubclassProcessor] = None
    ) -> DXFNamespace:
        dxf = super().load_dxf_attribs(processor)
        if processor:
            tags = processor.subclass_by_index(2)

            if tags:
                tags = Tags(self.load_boundary_path(tags))
                processor.fast_load_dxfattribs(
                    dxf, acdb_underlay_group_codes, subclass=tags
                )
                if len(self.boundary_path) < 2:
                    self.dxf = dxf
                    self.reset_boundary_path()
            else:
                raise const.DXFStructureError(
                    f"missing 'AcDbUnderlayReference' subclass in "
                    f"{self.DXFTYPE}(#{dxf.handle})"
                )
        return dxf

    def load_boundary_path(self, tags: Tags) -> Iterable:
        path = []
        for tag in tags:
            if tag.code == 11:
                path.append(tag.value)
            else:
                yield tag
        self._boundary_path = path

    def post_load_hook(self, doc: Drawing) -> None:
        super().post_load_hook(doc)
        db = doc.entitydb
        self._underlay_def = db.get(self.dxf.get("underlay_def_handle", None))  # type: ignore

    def post_bind_hook(self):
        assert isinstance(self.dxf.handle, str)
        underlay_def = self._underlay_def
        if (
            isinstance(underlay_def, UnderlayDefinition)
            and self.doc is underlay_def.doc  # it's not a xref copy!
        ):
            underlay_def.append_reactor_handle(self.dxf.handle)

    def export_entity(self, tagwriter: AbstractTagWriter) -> None:
        """Export entity specific data as DXF tags."""
        super().export_entity(tagwriter)
        tagwriter.write_tag2(SUBCLASS_MARKER, acdb_underlay.name)
        self.dxf.export_dxf_attribs(
            tagwriter,
            [
                "underlay_def_handle",
                "insert",
                "scale_x",
                "scale_y",
                "scale_z",
                "rotation",
                "extrusion",
                "flags",
                "contrast",
                "fade",
            ],
        )
        self.export_boundary_path(tagwriter)

    def export_boundary_path(self, tagwriter: AbstractTagWriter):
        for vertex in self.boundary_path:
            tagwriter.write_vertex(11, vertex[:2])

    def register_resources(self, registry: xref.Registry) -> None:
        super().register_resources(registry)
        if isinstance(self._underlay_def, UnderlayDefinition):
            registry.add_handle(self._underlay_def.dxf.handle)

    def map_resources(self, clone: Self, mapping: xref.ResourceMapper) -> None:
        assert isinstance(clone, Underlay)
        super().map_resources(clone, mapping)
        underlay_def_copy = self.map_underlay_def(clone, mapping)
        clone._underlay_def = underlay_def_copy
        clone.dxf.underlay_def_handle = underlay_def_copy.dxf.handle
        underlay_def_copy.append_reactor_handle(clone.dxf.handle)

    def map_underlay_def(
        self, clone: Underlay, mapping: xref.ResourceMapper
    ) -> UnderlayDefinition:
        underlay_def = self._underlay_def
        assert isinstance(underlay_def, UnderlayDefinition)

        underlay_def_copy = mapping.get_reference_of_copy(underlay_def.dxf.handle)
        assert isinstance(underlay_def_copy, UnderlayDefinition)

        doc = clone.doc
        assert doc is not None

        underlay_dict = doc.rootdict.get_required_dict(underlay_def.acad_dict_name)
        if underlay_dict.find_key(underlay_def_copy):  # entry already exist
            return underlay_def_copy

        # create required dictionary entry
        key = doc.objects.next_underlay_key(lambda k: k not in underlay_dict)
        underlay_dict.take_ownership(key, underlay_def_copy)
        return underlay_def_copy

    def set_underlay_def(self, underlay_def: UnderlayDefinition) -> None:
        self._underlay_def = underlay_def
        self.dxf.underlay_def_handle = underlay_def.dxf.handle
        underlay_def.append_reactor_handle(self.dxf.handle)

    def get_underlay_def(self) -> Optional[UnderlayDefinition]:
        return self._underlay_def

    @property
    def boundary_path(self):
        return self._boundary_path

    @boundary_path.setter
    def boundary_path(self, vertices: Iterable[UVec]) -> None:
        self.set_boundary_path(vertices)

    @property
    def clipping(self) -> bool:
        return bool(self.dxf.flags & const.UNDERLAY_CLIPPING)

    @clipping.setter
    def clipping(self, state: bool) -> None:
        self.set_flag_state(const.UNDERLAY_CLIPPING, state)

    @property
    def on(self) -> bool:
        return bool(self.dxf.flags & const.UNDERLAY_ON)

    @on.setter
    def on(self, state: bool) -> None:
        self.set_flag_state(const.UNDERLAY_ON, state)

    @property
    def monochrome(self) -> bool:
        return bool(self.dxf.flags & const.UNDERLAY_MONOCHROME)

    @monochrome.setter
    def monochrome(self, state: bool) -> None:
        self.set_flag_state(const.UNDERLAY_MONOCHROME, state)

    @property
    def adjust_for_background(self) -> bool:
        return bool(self.dxf.flags & const.UNDERLAY_ADJUST_FOR_BG)

    @adjust_for_background.setter
    def adjust_for_background(self, state: bool):
        self.set_flag_state(const.UNDERLAY_ADJUST_FOR_BG, state)

    @property
    def scaling(self) -> tuple[float, float, float]:
        return self.dxf.scale_x, self.dxf.scale_y, self.dxf.scale_z

    @scaling.setter
    def scaling(self, scale: Union[float, tuple]):
        if isinstance(scale, (float, int)):
            x, y, z = scale, scale, scale
        else:
            x, y, z = scale
        self.dxf.scale_x = x
        self.dxf.scale_y = y
        self.dxf.scale_z = z

    def set_boundary_path(self, vertices: Iterable[UVec]) -> None:
        # path coordinates as drawing coordinates but unscaled
        vertices = list(vertices)
        if len(vertices):
            self._boundary_path = vertices
            self.clipping = True
        else:
            self.reset_boundary_path()

    def reset_boundary_path(self) -> None:
        """Removes the clipping path."""
        self._boundary_path = []
        self.clipping = False

    def destroy(self) -> None:
        if not self.is_alive:
            return

        if self._underlay_def:
            self._underlay_def.discard_reactor_handle(self.dxf.handle)
        del self._boundary_path
        super().destroy()

    def transform(self, m: Matrix44) -> Underlay:
        """Transform UNDERLAY entity by transformation matrix `m` inplace.

        Unlike the transformation matrix `m`, the UNDERLAY entity can not
        represent a non-orthogonal target coordinate system and an
        :class:`InsertTransformationError` will be raised in that case.

        """
        dxf = self.dxf
        source_system = InsertCoordinateSystem(
            insert=Vec3(dxf.insert),
            scale=(dxf.scale_x, dxf.scale_y, dxf.scale_z),
            rotation=dxf.rotation,
            extrusion=dxf.extrusion,
        )
        try:
            target_system = source_system.transform(m)
        except InsertTransformationError:
            raise InsertTransformationError(
                "UNDERLAY entity can not represent a non-orthogonal target coordinate system."
            )
        dxf.insert = target_system.insert
        dxf.rotation = target_system.rotation
        dxf.extrusion = target_system.extrusion
        dxf.scale_x = target_system.scale_factor_x
        dxf.scale_y = target_system.scale_factor_y
        dxf.scale_z = target_system.scale_factor_z
        self.post_transform(m)
        return self


@register_entity
class PdfUnderlay(Underlay):
    """DXF PDFUNDERLAY entity"""

    DXFTYPE = "PDFUNDERLAY"


@register_entity
class PdfReference(Underlay):
    """PDFREFERENCE ia a synonym for PDFUNDERLAY, ezdxf creates always PDFUNDERLAY
    entities.
    """

    DXFTYPE = "PDFREFERENCE"


@register_entity
class DwfUnderlay(Underlay):
    """DXF DWFUNDERLAY entity"""

    DXFTYPE = "DWFUNDERLAY"


@register_entity
class DgnUnderlay(Underlay):
    """DXF DGNUNDERLAY entity"""

    DXFTYPE = "DGNUNDERLAY"


acdb_underlay_def = DefSubclass(
    "AcDbUnderlayDefinition",
    {
        "filename": DXFAttr(1),  # File name of underlay
        "name": DXFAttr(2),
        # underlay name - pdf=page number to display; dgn=default; dwf=????
    },
)
acdb_underlay_def_group_codes = group_code_mapping(acdb_underlay_def)


# (PDF|DWF|DGN)DEFINITION - requires entry in objects table ACAD_(PDF|DWF|DGN)DEFINITIONS,
# ACAD_(PDF|DWF|DGN)DEFINITIONS do not exist by default
class UnderlayDefinition(DXFObject):
    """Virtual UNDERLAY DEFINITION entity."""

    DXFTYPE = "UNDERLAYDEFINITION"
    DXFATTRIBS = DXFAttributes(base_class, acdb_underlay_def)
    MIN_DXF_VERSION_FOR_EXPORT = DXF2000

    def load_dxf_attribs(
        self, processor: Optional[SubclassProcessor] = None
    ) -> DXFNamespace:
        dxf = super().load_dxf_attribs(processor)
        if processor:
            processor.fast_load_dxfattribs(
                dxf, acdb_underlay_def_group_codes, subclass=1
            )
        return dxf

    def export_entity(self, tagwriter: AbstractTagWriter) -> None:
        """Export entity specific data as DXF tags."""
        super().export_entity(tagwriter)
        tagwriter.write_tag2(SUBCLASS_MARKER, acdb_underlay_def.name)
        self.dxf.export_dxf_attribs(tagwriter, ["filename", "name"])

    @property
    def file_format(self) -> str:
        return self.DXFTYPE[:3]

    @property
    def entity_name(self) -> str:
        return self.file_format + "UNDERLAY"

    @property
    def acad_dict_name(self) -> str:
        return f"ACAD_{self.file_format}DEFINITIONS"

    def post_new_hook(self):
        self.set_reactors([self.dxf.owner])


@register_entity
class PdfDefinition(UnderlayDefinition):
    """DXF PDFDEFINITION entity"""

    DXFTYPE = "PDFDEFINITION"


@register_entity
class DwfDefinition(UnderlayDefinition):
    """DXF DWFDEFINITION entity"""

    DXFTYPE = "DWFDEFINITION"


@register_entity
class DgnDefinition(UnderlayDefinition):
    """DXF DGNDEFINITION entity"""

    DXFTYPE = "DGNDEFINITION"



# Copyright (c) 2018-2020 Manfred Moitzi
# License: MIT License
from ezdxf.math import PassTroughUCS, UCS, Vec3, X_AXIS, Y_AXIS, Z_AXIS


def test_u_vectors():
    ucs = PassTroughUCS()

    assert ucs.ux == X_AXIS
    assert ucs.uy == Y_AXIS
    assert ucs.uz == Z_AXIS


def test_to_wcs():
    ucs = PassTroughUCS()
    assert ucs.to_wcs((1, 2, 3)) == Vec3(1, 2, 3)
    assert ucs.to_wcs((3, 4, 5)) == Vec3(3, 4, 5)

    ucs2 = UCS()
    assert ucs.to_wcs((1, 2, 3)) == ucs2.to_wcs((1, 2, 3))
    assert ucs.to_wcs((3, 4, 5)) == ucs2.to_wcs((3, 4, 5))


def test_points_to_wcs():
    ucs = PassTroughUCS()
    assert list(ucs.points_to_wcs([(1, 2, 3), (3, 4, 5)])) == [
        Vec3(1, 2, 3),
        Vec3(3, 4, 5),
    ]

    ucs2 = UCS()
    assert list(ucs.points_to_wcs([(1, 2, 3), (3, 4, 5)])) == list(
        ucs2.points_to_wcs([(1, 2, 3), (3, 4, 5)])
    )


def test_to_ocs():
    ucs = PassTroughUCS()
    assert ucs.to_ocs((1, 2, 3)) == Vec3(1, 2, 3)
    assert ucs.to_ocs((3, 4, 5)) == Vec3(3, 4, 5)


def test_points_to_ocs():
    ucs = PassTroughUCS()
    assert list(ucs.points_to_ocs([(1, 2, 3), (3, 4, 5)])) == [
        Vec3(1, 2, 3),
        Vec3(3, 4, 5),
    ]


def test_from_wcs():
    ucs = PassTroughUCS()
    assert ucs.from_wcs((1, 2, 3)) == Vec3(1, 2, 3)
    assert ucs.from_wcs((3, 4, 5)) == Vec3(3, 4, 5)

    ucs2 = UCS()
    assert ucs.from_wcs((1, 2, 3)) == ucs2.from_wcs(Vec3(1, 2, 3))
    assert ucs.from_wcs((3, 4, 5)) == ucs2.from_wcs(Vec3(3, 4, 5))


def test_points_from_wcs():
    points = Vec3.list([(1, 2, 3), (3, 4, 5)])
    ucs = PassTroughUCS()
    assert list(ucs.points_from_wcs(points)) == points

    ucs2 = UCS()
    assert list(ucs.points_from_wcs(points)) == list(
        ucs2.points_from_wcs(points)
    )
